#ifndef JELIZA
#define JELIZA 1


/*
 * This is part of JEliza revision 5X.
 * Copyright 2008 by Tobias Schulz
 * WWW: http://jeliza.ch.to/
 *
 * JEliza is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later
 * version.
 *
 * JEliza is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU GPL
 * along with JEliza (file COPYING) ; if not, write
 * to the Free Software Foundation, Inc., 51 Franklin St,
 * Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include "defs.h"

using namespace std;
using jeliza::log;

/*
 * Benutze die Online DB?
 */
bool jeliza::use_online_db = true;
jeliza::string online_db_new_entry_url("http://jeliza.kilu.de/developer/online-db/new.php?");
jeliza::string online_db_new_sentence_entry_url("http://jeliza.kilu.de/developer/online-db/new_sentence.php?");

/*
 * Ein besseres "pair"-Objekt als in der Standardbibliothek
 */
namespace jeliza {
    template <typename T1, typename T2>
    class pair {
    public:
        T1 first;
        T2 second;

        pair (T1 f) {
            first = f;
        }
        pair (T2 s) {
            second = s;
        }
        pair& operator= (const pair& p) {
            first = p.first;
            second = p.second;

            return (*this);
        }
        bool operator== (const pair& p) {
            return (first == p.first && second == p.second);
        }
    };
}


/*
 * Logdatei oeffnen
 *
 * ofstream - output file stream
 */
ofstream clogger("jeliza.log");

/*
 * Ist die Python-Initialisierung abgeschlossen?
 */
bool jeliza::is_python_initialized = false;

/*
 * Damit es ein Objekt gibt für die spo_split-Methoden
 *
 * Dies ist ein Singleton!
 */
jeliza::spo_split jeliza::spo_split::instance;

/*
 * Auf eine bestimmte answer bestehen, in "Notfaellen"?
 */
bool jeliza::force_answer_check = false;
jeliza::string jeliza::force_answer;
jeliza::string jeliza::force_answer_question;
string jeliza::force_answer_word;
string jeliza::force_answer_genus;
string jeliza::force_noun_or_not;

/*
 * Globale Initialisierung abgeschlossen?
 */
bool jeliza::global_fertig = false;

/*
 * Globale JEliza-Implementierung
 */
jeliza::jeliza_impl jeliza_impl_instance;

/*
 * Naechste wordtype von GUI
 */
std::string jeliza::next_word_type("-1");
std::string jeliza::next_genus("-1");
std::string jeliza::next_noun_or_not("-1");

map<std::string, jeliza::wordtype> cache_noun_or_not;

/*
 * Globale Initialisierungs-Funktion
 */
void jeliza::jeliza_impl::init_databases() {
    log("Loading brain...");

    long max_size = 210000;

    /*
     * Datenbank leeren
     */
    if (jeliza::spo_split::get().wordtype_database != NULL)
        delete jeliza::spo_split::get().wordtype_database;
    jeliza::spo_split::get().wordtype_database = new vector<jeliza::strings>();
    jeliza::spo_split::get().wordtype_equal.reset(new vector<jeliza::pair<std::string, jeliza::string> >());
    jeliza::spo_split::get().wordtype_ohne_prefix.reset(new vector<jeliza::pair<std::string, jeliza::string> >());
    jeliza::spo_split::get().wordtype_equal_low.reset(new vector<jeliza::pair<std::string, jeliza::string> >());
    jeliza::spo_split::get().wordtype_ohne_prefix_low.reset(new vector<jeliza::pair<std::string, jeliza::string> >());
    jeliza::spo_split::get().genus_index.reset(new map<std::string, std::string>);
    jeliza::spo_split::get().genuss.reset(new vector<std::string>);

    /*
     * Benoetigten Speicherplatz ausrechnen und reservieren
     */
    long long required_space = 5 * 2 * max_size * sizeof(jeliza::pair<std::string, jeliza::string>);
    std::stringstream sst;
    std::string req;
    sst << required_space / 1024 / 1024;
    sst >> req;
    log("Reserving space for brain: " + req + " MiB");

    jeliza::spo_split::get().wordtype_database->reserve(max_size * 2);
    jeliza::spo_split::get().wordtype_equal->reserve(max_size * 2);
    jeliza::spo_split::get().wordtype_equal_low->reserve(max_size * 2);
    jeliza::spo_split::get().wordtype_ohne_prefix->reserve(max_size * 2);
    jeliza::spo_split::get().wordtype_ohne_prefix_low->reserve(max_size * 2);

    /*
     * Datenbank einlesen
     */
    log("Lade Gehirn aus Datei...");
    ifstream iDic(("lang_" + jeliza::get_lang_str() + "/wsh/dic.wsh").ref().c_str());
    jeliza::string str;
    int k = 0;
    unsigned int procent;
    while (iDic) {
        ++k;
        if (k % 3000 == 0) {
            procent = (int)(100.0 / 90000.0 * k);
            cout << "\r" << procent << "% geladen                                            \r" << flush;
            jeliza::update_prozent(procent);
        }

        getline(iDic, str.ref());
        jeliza::strings vec;

        util::split(str, std::string(";"), vec);
        if (vec.size() > 2) {
            add_wortschatz_word(vec);
        }
    }

    cout << endl;

    ifstream igenus(("lang_" + jeliza::get_lang_str() + "/wsh/genus.wsh").ref().c_str());
    while (igenus) {
        ++k;
        if (k % 3000 == 0) {
            procent = (unsigned int)(100.0 / 90000.0 * (double)(k));
            cout << "\r" << procent << "% geladen                                            \r" << flush;
            jeliza::update_prozent(procent);
        }

        getline(igenus, str.ref());
        jeliza::strings vec;

        util::split(str, std::string(";"), vec);
        if (vec.size() == 2) {
            add_wortschatz_genus(vec);
        }
    }

    /*
     * Satzauftrenndatenbank initialisieren
     */
    jeliza::spo_split::get().init_sentences_splitted();

    /*
     * Wissensdatenbank initialiseren
     */
    init();
}

void jeliza::jeliza_impl::start() {
    jeliza_impl_instance = *this;

    /*
     * Python und Python-Threads aktivieren, das C++-Modul registrieren
     */
    Py_Initialize();
    PyEval_InitThreads();
    PyImport_AddModule("jelizacpp");
    Py_InitModule("jelizacpp", jelizacpp_methods);

    /*
     * Python ist initialisiert
     */
    jeliza::is_python_initialized = true;

    /*
     * Datenbanken initialiseren
     */
    init_databases();

    /*
     * Server starten
     */
    PyRun_SimpleString("execfile(\"modules/_server.py\")");

    jeliza::global_fertig = true;
}

/*
 * Ein Wort in die Wortschatz-Datenbank eintragen
 */
void jeliza::jeliza_impl::add_wortschatz_word(jeliza::strings vec) {
    vec[0].lower();
    vec[2].strip();
    vec[3].strip();

    jeliza::strings vec_tmp;

    jeliza::spo_split::get().wordtype_database->push_back(vec);
    jeliza::spo_split::get().wordtype_equal->push_back(jeliza::pair<std::string, jeliza::string>(vec[1].ref()));
    jeliza::spo_split::get().wordtype_equal->push_back(jeliza::pair<std::string, jeliza::string>(vec[0]));
    jeliza::spo_split::get().wordtype_ohne_prefix->push_back(jeliza::pair<std::string, jeliza::string>(vec[3].ref()));
    jeliza::spo_split::get().wordtype_ohne_prefix->push_back(jeliza::pair<std::string, jeliza::string>(vec[0]));

    jeliza::spo_split::get().wordtype_equal_low->push_back(jeliza::pair<std::string, jeliza::string>(vec[1].get_lower().ref()));
    jeliza::spo_split::get().wordtype_equal_low->push_back(jeliza::pair<std::string, jeliza::string>(vec[0]));
    jeliza::spo_split::get().wordtype_ohne_prefix_low->push_back(jeliza::pair<std::string, jeliza::string>(vec[3].get_lower().ref()));
    jeliza::spo_split::get().wordtype_ohne_prefix_low->push_back(jeliza::pair<std::string, jeliza::string>(vec[0]));
}

void jeliza::jeliza_impl::add_wortschatz_genus(jeliza::strings vec) {
    vec[0].lower();
    vec[1].strip();

    jeliza::strings vec_tmp;

    (*jeliza::spo_split::get().genus_index)[vec[1].ref()] = vec[0].ref();
    jeliza::spo_split::get().genuss->push_back(vec[1].ref());
}

/*
 * Ein Wort in die Datenbank eintragen, nach erzwungener answer
 */
bool jeliza::jeliza_impl::add_word(char ch) {
    jeliza::string to_write = jeliza::string(";" + jeliza::force_answer_word + ";" + spo_split::get().detect_word_base(jeliza::string(jeliza::force_answer_word)).ref() + ";" + spo_split::get().delete_prefixes(jeliza::string(jeliza::force_answer_word)).ref());
    unsigned int size = to_write.size();
    switch (ch) {
        case '1':
            to_write = "vt" + to_write;
            break;
        case '2':
            to_write = "n," + to_write;
            break;
        case '3':
            to_write = "adj" + to_write;
            break;
        case '4':
            to_write = "n," + to_write;
            break;
        case '5':
            to_write = "fw" + to_write;
            break;
        case '6':
            to_write = "prep" + to_write;
            break;
        case '7':
            to_write = "inter" + to_write;
            break;
    }

    if (size == to_write.size()) {
        return false;
    }

    jeliza::strings vec;

    util::split(to_write, std::string(";"), vec);
    while (vec.size() < 4) {
        vec.push_back(s("_"));
    }
    add_wortschatz_word(vec);

    jeliza::force_answer_check = false;
    jeliza::force_answer_word = "";
    jeliza::force_answer.set("");

    return true;
}

/*
 * Ein genus in die Datenbank eintragen, nach erzwungener answer
 */
bool jeliza::jeliza_impl::add_genus(char ch, std::string genus) {
    jeliza::string to_write = jeliza::string(";" + genus);
    unsigned int size = to_write.size();
    switch (ch) {
        case '1':
            to_write = "male" + to_write;
            break;
        case '2':
            to_write = "female" + to_write;
            break;
        case '3':
            to_write = "thing" + to_write;
            break;
    }

    if (size == to_write.size()) {
        return false;
    }

    jeliza::strings vec;

    util::split(to_write, std::string(";"), vec);
    add_wortschatz_genus(vec);

    jeliza::force_answer_genus = "";

    return true;
}

/*
 * Wrapper fuer std::find
 */
jeliza::strings jeliza::findit(shared_ptr<vector<jeliza::pair<std::string, jeliza::string> > >* vec, std::string str) {
    vector<jeliza::pair<std::string, jeliza::string> >::iterator it = find((*vec)->begin(), (*vec)->end(), jeliza::pair<std::string, jeliza::string>(str));
    if (it == (*vec)->end()) {
        return jeliza::strings();
    }

    ++it;
    jeliza::strings ret;
    ret.push_back(it->second);
    return ret;
}

/*
 * Encodiere URLs
 */
jeliza::string url_encode(jeliza::string val)
{
  jeliza::string ret(val);

  ret = ret.replace_nocase_safe(s("\r"), s(""))
           .replace_nocase_safe(s("\n"), s("NEWLINE"))
           .replace_nocase_safe(s(" "), s("%20"))
           .replace_nocase_safe(s("<"), s("%3C"))
           .replace_nocase_safe(s(">"), s("%3E"))
           .replace_nocase_safe(s("#"), s("%23"))
           .replace_nocase_safe(s("%"), s("%25"))
           .replace_nocase_safe(s("{"), s("%7B"))
           .replace_nocase_safe(s("}"), s("%7D"))
           .replace_nocase_safe(s("|"), s("%7C"))
           .replace_nocase_safe(s("\\"), s("%5C"))
           .replace_nocase_safe(s("^"), s("%5E"))
           .replace_nocase_safe(s("~"), s("%7E"))
           .replace_nocase_safe(s("["), s("%5B"))
           .replace_nocase_safe(s("]"), s("%5D"))
           .replace_nocase_safe(s("`"), s("%60"))
           .replace_nocase_safe(s(";"), s("%3B"))
           .replace_nocase_safe(s("/"), s("%2F"))
           .replace_nocase_safe(s("?"), s("%3F"))
           .replace_nocase_safe(s(":"), s("%3A"))
           .replace_nocase_safe(s("@"), s("%40"))
           .replace_nocase_safe(s("="), s("%3D"))
           .replace_nocase_safe(s("&"), s("%26"))
           .replace_nocase_safe(s("$"), s("%24"));

  return ret;
}


/*
 * In XML umwandeln und zurueckgeben
 */
jeliza::string jeliza::db_sentence::to_xml() {
    jeliza::string temp;

    std::stringstream sst;
    sst << priority;
    jeliza::string tempprio;
    sst >> tempprio.ref();

    jeliza::string all;

    all += " <fact>\n";
    all += this->print_part (string("prefix>   "), temp, prefix.ref());
    all += this->print_part (string("subject>  "), temp, subject.ref());
    all += this->print_part (string("verb>     "), temp, verb.ref());
    all += this->print_part (string("object>   "), temp, object.ref());
    all += this->print_part (string("suffix>   "), temp, suffix.ref());
    all += this->print_part (string("feeling>  "), temp, feeling.ref());
    all += this->print_part (string("priority> "), temp, tempprio.ref());
    all += this->print_part (string("category> "), temp, category.ref());
    all += this->print_part (string("additional>   "), temp, additional.ref());
    all += this->print_part (string("quesword> "), temp, quesword.ref());
    for (unsigned int x = 0; x < sub->size(); x++) {
        all += " <sub>\n";
        all += (*sub)[x].to_xml();
        all += " </sub>\n";
    }
    all += " </fact>\n\n";

    return all;
}

bool jeliza::db_sentence::operator== (const db_sentence& sent) const {
    if (prefix == sent.prefix
            && subject == sent.subject
            && verb == sent.verb
            && object == sent.object
            && suffix == sent.suffix
            && feeling == sent.feeling
            && additional == sent.additional
            && quesword == sent.quesword
            ) {
        return true;
    }
    return false;
}


/*
 * In XML umwandeln und in Log(-Datei?) schreiben
 */
void jeliza::db_sentence::to_xml_print() {
    clogger << to_xml().ref() << endl;
}

/*
 * Hilfsfunktion fuer to_xml
 */
jeliza::string jeliza::db_sentence::print_part (string str, jeliza::string temp, std::string wert) {
    temp = temp.replace(string("\n"), std::string("\\n"));
    temp = temp.replace(string("\r"), std::string(""));

    return jeliza::string("  <" + str + wert + "  " + "</" + str + "\n");
}

/*
 * Den Satz aufgetrennt menschenlesbar ausgeben
 */
void jeliza::db_sentence::print(std::string begin = "", bool clog = true, bool cou = true) {
    if (clog) {
        clogger << begin << "Subject:   " << subject.ref() << endl;
        clogger << begin << "Verb:      " << verb.ref() << endl;
        clogger << begin << "Object:    " << object.ref() << endl;
        clogger << begin << "Prefix:    " << prefix.ref() << endl;
        clogger << begin << "Suffix:    " << suffix.ref() << endl;
        clogger << begin << "Feeling:   " << feeling.ref() << endl;
        clogger << begin << "Priority:  " << priority << endl;
        clogger << begin << "Category:  " << category.ref() << endl;
        clogger << begin << "additional:    " << additional.ref() << endl;
        clogger << begin << "QuesWord:  " << quesword.ref() << endl;
        for (unsigned int x = 0; x < sub->size(); x++) {
            clogger << begin << "Nebensatz " << x+1 << ":" << endl;
            (*sub)[x].print(begin + "  ", true, false);
        }
        clogger << endl;
    }

    if (cou) {
        cout << begin << "Subject:   " << subject.ref() << endl;
        cout << begin << "Verb:      " << verb.ref() << endl;
        cout << begin << "Object:    " << object.ref() << endl;
        cout << begin << "Prefix:    " << prefix.ref() << endl;
        cout << begin << "Suffix:    " << suffix.ref() << endl;
        cout << begin << "Feeling:   " << feeling.ref() << endl;
        cout << begin << "Priority:  " << priority << endl;
        cout << begin << "Category:  " << category.ref() << endl;
        cout << begin << "additional:    " << additional.ref() << endl;
        cout << begin << "QuesWord:  " << quesword.ref() << endl;
        for (unsigned int x = 0; x < sub->size(); x++) {
            cout << begin << "Nebensatz " << x+1 << ":" << endl;
            (*sub)[x].print(begin + "  ", false, true);
        }
        cout << endl;
    }
}

/*
 * Leerzeichen am Anfang und Ende von allen sentence_parten entfernen
 */
void jeliza::db_sentence::strip() {
    prefix.strip();
    subject.strip();
    verb.strip();
    object.strip();
    suffix.strip();
    feeling.strip();
    category.strip();
    additional.strip();
    quesword.strip();

    prefix.replace(string("\\n"), std::string("\n"));
    subject.replace(string("\\n"), std::string("\n"));
    verb.replace(string("\\n"), std::string("\n"));;
    object.replace(string("\\n"), std::string("\n"));
    suffix.replace(string("\\n"), std::string("\n"));
    feeling.replace(string("\\n"), std::string("\n"));
    category.replace(string("\\n"), std::string("\n"));
    additional.replace(string("\\n"), std::string("\n"));
    quesword.replace(string("\\n"), std::string("\n"));

    for (unsigned int x = 0; x < sub->size(); x++) {
        (*sub)[x].strip();
    }
}

jeliza::strings jeliza::db_sentence::gen_sentences(bool withFix = false) {
    strip();

    jeliza::strings ans;

    if (subject.size() > 2 && verb.size() > 2 && object.size() > 1) {
        ans.push_back(jeliza::string(subject + " " + verb + " " + object));
    } else if (withFix) {
        ans.push_back(jeliza::string(prefix + " " + subject + " " + verb + " " + object + " " + suffix));
    }
    jeliza::string temp;
    (temp = prefix + " " + subject + " " + verb + " " + object).strip();
    if (temp.size() > 1) {
        ans.push_back(temp);
    }
    (temp = subject + " " + verb + " " + object + " " + suffix).strip();
    if (temp.size() > 1) {
        ans.push_back(temp);
    }

    ans.push_back(jeliza::string(prefix + " " + subject + " " + verb + " " + object + " " + suffix));

    return ans;
}

jeliza::strings jeliza::db_sentence::gen_sentences_all(bool) {
    strip();

    jeliza::strings ans;

    ans.push_back(jeliza::string(prefix + " " + subject + " " + verb + " " + object + " " + suffix));

    return ans;
}

/*
 * Ist es eine question?
 */
unsigned int jeliza::jeliza_impl::is_question (jeliza::string ques) {
    ques = " " + ques + " ";
    ques.lower();
    if (ques.contains(jeliza::string(" why "))) {
        return 1;
    }
    if (ques.contains(jeliza::string(" when "))) {
        return 1;
    }
    if (ques.contains(jeliza::string(" who "))) {
        return 1;
    }
    if (ques.contains(jeliza::string(" where "))) {
        return 1;
    }
    if (ques.contains(jeliza::string(" what "))) {
        return 1;
    }
    if (ques.contains(jeliza::string(" which "))) {
        return 1;
    }
    if (ques.contains(jeliza::string(" was "))) {
        return 1;
    }
    if (ques.contains(jeliza::string(" wie "))) {
        return 1;
    }
    if (ques.contains(jeliza::string(" wer "))) {
        return 1;
    }
    if (ques.contains(jeliza::string(" wie "))) {
        return 1;
    }
    if (ques.contains(jeliza::string(" wo "))) {
        return 1;
    }
    if (ques.contains(jeliza::string(" wann "))) {
        return 1;
    }
    if (ques.contains(jeliza::string(" warum "))) {
        return 1;
    }
    if (ques.contains(jeliza::string(" wieso "))) {
        return 1;
    }
    if (ques.contains(jeliza::string(" weshalb "))) {
        return 1;
    }

    if (ques.contains(jeliza::string(" kennst du ")) && ques.contains(jeliza::string("?"))) {
        return 1;
    }

    if (ques.contains(jeliza::string(" kenne du ")) && ques.contains(jeliza::string("?"))) {
        return 1;
    }

    if (ques.contains(jeliza::string("?"))) {
        return 2;
    }

    return 0;
}

jeliza::db_sentence& jeliza::db_sentence::operator= (const jeliza::db_sentence& sent) {
    return set(sent);
}

jeliza::db_sentence::db_sentence (const jeliza::db_sentence& sent) {
    set(sent);
}

jeliza::db_sentence& jeliza::db_sentence::set (const jeliza::db_sentence& sent) {
    subject.set(sent.subject);
    verb.set(sent.verb);
    object.set(sent.object);
    prefix.set(sent.prefix);
    suffix.set(sent.suffix);
    feeling.set(sent.feeling);
    category.set(sent.category);
    quesword.set(sent.quesword);
    additional.set(sent.additional);
    priority = sent.priority;
    sub.reset(new vector<jeliza::db_sentence>());

    for (unsigned int x = 0; x < sent.sub->size(); x++) {
        db_sentence dbs = (*sent.sub)[x];
        sub->push_back(dbs);
    }

    return (*this);
}

/*
 * Datenbank aus Datei einlesen
 */
jeliza::DB* jeliza::db_manager::parse (std::string file) {
    ifstream i(file.c_str());

    jeliza::string buffer;
    jeliza::string all;
    while (i) {
        getline (i, buffer.ref());

        if (buffer.contains("<?xml")) {
            continue;
        }

        all += buffer;
        all += "\n";
    }

    jeliza::strings lines;
    util::split(all, std::string("\n\r<>"), lines);

    jeliza::DB* sents = new jeliza::DB;

    for (jeliza::strings::iterator it = lines.begin(); it != lines.end(); it++) {
        it->strip();
    }

    jeliza::db_sentence act_sent;
    jeliza::db_sentence act_sent_vor_sub;
    jeliza::string line;
    bool sub_mode = false;
    int a = 0;
    for (jeliza::strings::iterator it = lines.begin(); it != lines.end(); it++) {
        a++;
        line.ref() = it->ref();

        jeliza::update_prozent(100.0 / lines.size() * a);

        if (line.equal("priority")) {
            std::stringstream sst;
            sst << (*(it + 1)).ref();
            int t;
            sst >> t;
            act_sent.priority = t;
            it += 2;
        }

        else if (line.equal("subject")) {
            act_sent.subject.ref() = (it + 1)->get();

            it += 2;
        }

        else if (line.equal("verb")) {
            act_sent.verb.ref() = (it + 1)->get();
            it += 2;
        }

        else if (line.equal("object")) {
            act_sent.object.ref() = (it + 1)->get();
            it += 2;
        }

        else if (line.equal("feeling")) {
            act_sent.feeling.ref() = (it + 1)->get();
            it += 2;
        }

        else if (line.equal("prefix")) {
            act_sent.prefix.ref() = (it + 1)->get();
            it += 2;
        }

        else if (line.equal("suffix")) {
            act_sent.suffix.ref() = (it + 1)->get();
            it += 2;
        }

        else if (line.equal("category")) {
            act_sent.category.ref() = (it + 1)->get();
            it += 2;
        }

        else if (line.equal("additional")) {
            act_sent.additional.ref() = (it + 1)->get();
            it += 2;
        }

        else if (line.equal("quesword")) {
            act_sent.quesword.ref() = (it + 1)->get();
            it += 2;
        }

        else if (line.equal("/fact")) {
            if (!sub_mode) {
                act_sent.strip();
                if (jeliza::jeliza_impl::is_question(act_sent.gen_sentences_all(true)[0]) == 0) {
                    sents->push_back(act_sent);
                } else {
                    cout << "Ist question: " << act_sent.gen_sentences_all(true)[0].ref() << endl;
                }
                act_sent = jeliza::db_sentence();
            }
        }

        else if (line.equal("/sub")) {
            act_sent.strip();
            act_sent_vor_sub.sub->push_back(act_sent);

            act_sent = act_sent_vor_sub;
            sub_mode = false;
        }

        else if (line.equal("sub")) {
            act_sent_vor_sub = act_sent;
            act_sent = jeliza::db_sentence();
            sub_mode = true;
        }



    }

    return (sents);
}

void berichtige_wenn_dann(jeliza::db_sentence act, jeliza::db_sentence sub) {
    if (sub.quesword.get_lower().contains("wenn") || sub.quesword.get_lower().contains("falls")) {
        sub.quesword.set("");
    }

    if (act.quesword.get_lower().contains("wenn") || act.quesword.get_lower().contains("falls")) {
        act.quesword.set("");

        jeliza::db_sentence tmp;
        tmp = act;
        act = sub;
        sub = tmp;
    }
}

/*
 * Wrapper fuer
 * Trennt den Satz auf
 */
shared_ptr<jeliza::db_sentence> jeliza::spo_split::get_spo (jeliza::string buffer) {
    buffer.replace(jeliza::string("\n"), jeliza::string("\\n"));

//    log("--- Split Sentence ---");
    std::pair<vector<jeliza::db_sentence>, vector<jeliza::db_sentence> > parts = split_spo(buffer);

    jeliza::db_sentence act_sent;

    if (!parts.first.size()) {
        if (parts.second.size()) {
            vector<db_sentence> questions;
            vector<db_sentence> subclauses;
            for (unsigned int x = 0; x < parts.second.size(); ++x) {
                if  (
                           parts.second[x].quesword.get_lower() == "wer"
                        or parts.second[x].quesword.get_lower() == "was"
                        or parts.second[x].quesword.get_lower() == "wem"
                        or parts.second[x].quesword.get_lower() == "wen"
                        or parts.second[x].quesword.get_lower() == "wie"
                        or parts.second[x].quesword.get_lower() == "warum"
                        or parts.second[x].quesword.get_lower() == "wieso"
                        or parts.second[x].quesword.get_lower() == "weshalb"
                        or parts.second[x].quesword.get_lower() == "wann"
                        or parts.second[x].quesword.get_lower() == "wo"
                        or parts.second[x].quesword.get_lower() == "where"
                        or parts.second[x].quesword.get_lower() == "who"
                        or parts.second[x].quesword.get_lower() == "when"
                        or parts.second[x].quesword.get_lower() == "how"
                        or parts.second[x].quesword.get_lower() == "whom"
                        or parts.second[x].quesword.get_lower() == "what"
                        or parts.second[x].quesword.get_lower() == "whether"
                        or parts.second[x].quesword.get_lower() == "why"
                    )
                {
                    questions.push_back(parts.second[x]);
                }
                else
                {
                    subclauses.push_back(parts.second[x]);
                }
            }
            parts.first = questions;
            parts.second = subclauses;
//            parts.first.push_back(parts.second[0]);
        }
    }

//    if (parts.size()) {
    if (parts.first.size()) {
//        cout << "Sentence Split Success" << endl;
        act_sent = parts.first[0];

        for (unsigned int x = 0; x < parts.second.size(); ++x) {
            act_sent.sub->push_back(parts.second[x]);
        }
//        act_sent.print();

        if (parts.second.size() >= 1
                && act_sent.subject.replace_nocase_safe(s("nothing"), s("")).replace_nocase_safe(s("{"), s("")).replace_nocase_safe(s("}"), s("")).get_strip().size() == 0) {
            act_sent = parts.second[0]; // questionsatz!
            for (unsigned int x = 1; x < parts.second.size(); ++x) {
                act_sent.sub->push_back(parts.second[x]);
            }
        }
    }

    shared_ptr<jeliza::db_sentence> act_sent_ret(new jeliza::db_sentence());
    (*act_sent_ret) = act_sent;

    return act_sent_ret;


/*
    jeliza::db_sentence sub = split_spo(buffer, SUB_CLAUSE);

    jeliza::db_sentence act_sent = parts;

    berichtige_wenn_dann(act_sent, sub);

    if (sub.verb.size() > 0) {
        act_sent.sub->push_back(sub);
    }
    cout << "act_sent.subject.replace_nocase_safe(s('nothing'), s("")).replace_nocase_safe(s('{'), s("")).replace_nocase_safe(s('}'), s('')).get_strip()"
         << act_sent.subject.replace_nocase_safe(s("nothing"), s("")).replace_nocase_safe(s("{"), s("")).replace_nocase_safe(s("}"), s("")).get_strip().ref()
         << endl;
    if (act_sent.subject.replace_nocase_safe(s("nothing"), s("")).replace_nocase_safe(s("{"), s("")).replace_nocase_safe(s("}"), s("")).get_strip().size() == 0) {
        act_sent = sub; // questionsatz!
    }

    shared_ptr<jeliza::db_sentence> act_sent_ret(new jeliza::db_sentence());
    (*act_sent_ret) = act_sent;

    return act_sent_ret;
*/
}


/*
 * Speichert die Datenbank
 */
void jeliza::db_manager::save(std::string file) {
    ofstream o(file.c_str());

    o << get_string() << endl;
    o.close();
}

/*
 * Liefert die Datenbank im XML-Format als std::string zurueck
 */
string jeliza::db_manager::get_string() {
    std::string str;

    str += "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>\n\n<jdb>\n";

    str += " <feelings>\n  <feeling>happy</happy>\n  <feeling>angry</angry>\n  <feeling>bored</bored>\n  <feeling>scared</scared>\n  <feeling>curiously</curiously>\n  <feeling>normal</normal>\n </feelings>\n";

    for (unsigned int x = 0; x < m_db->size(); ++x) {
        db_sentence s((*m_db)[x]);
        str += s.to_xml().toASCII().ref();
        str += "\n";
    }

    str += "</jdb>\n";

    return str;
}


/*
 * Vereinfacht einen Satz, z.B. zum Vergleichen
 */
jeliza::string jeliza::without_nonsense(jeliza::string question) {
	question = question.replace(string("?"), std::string(""));
	question = question.replace(string("!"), std::string(""));
	question = question.replace(string("."), std::string(""));
	question = question.replace(string(","), std::string(""));
	question = question.replace(string(";"), std::string(""));
	question = question.replace(string("+"), std::string(" ztrgftredrefd "));
	question = question.replace(string("ztrgftredrefd"), std::string("+"));
	question = question.replace(string("*"), std::string(" ztrgftredrefd "));
	question = question.replace(string("ztrgftredrefd"), std::string("*"));
	question = question.replace(string("/"), std::string(" ztrgftredrefd "));
	question = question.replace(string("ztrgftredrefd"), std::string("/"));
	question = question.replace(string("  "), std::string(" "));
	question = " " + question + " ";

	vector<std::string> unuseful_words;
	unuseful_words.push_back("der");
	unuseful_words.push_back("die");
	unuseful_words.push_back("das");
	unuseful_words.push_back("dem");
	unuseful_words.push_back("des");
	unuseful_words.push_back("dessen");
	unuseful_words.push_back("was");
	unuseful_words.push_back("wer");
	unuseful_words.push_back("wie");
	unuseful_words.push_back("wo");
	unuseful_words.push_back("wann");
	unuseful_words.push_back("von");
	unuseful_words.push_back("vom");
	unuseful_words.push_back("in");
	unuseful_words.push_back("auf");
	unuseful_words.push_back("unter");
	unuseful_words.push_back("neben");
	unuseful_words.push_back("zwischen");
	unuseful_words.push_back("denn");
	unuseful_words.push_back("bloß");
	unuseful_words.push_back("bloss");
	unuseful_words.push_back("stimmts");
	unuseful_words.push_back("stimmt's");
	unuseful_words.push_back("hmm");
	unuseful_words.push_back("hm");
	unuseful_words.push_back("hmmm");
	unuseful_words.push_back("hmmmm");
	unuseful_words.push_back("sonst");
	unuseful_words.push_back("ansonsten");
	unuseful_words.push_back("noch");

	unuseful_words.push_back("fast");
	unuseful_words.push_back("bald");
	unuseful_words.push_back("sehr");
	unuseful_words.push_back("na");
	unuseful_words.push_back("bisschen");
	unuseful_words.push_back("bischen");
	unuseful_words.push_back("hald");
	unuseful_words.push_back("halt");
	unuseful_words.push_back("eben");
	unuseful_words.push_back("oder");
	unuseful_words.push_back("und");

	unuseful_words.push_back("of");
	unuseful_words.push_back("from");
	unuseful_words.push_back("by");
	unuseful_words.push_back("between");
	unuseful_words.push_back("what");
	unuseful_words.push_back("when");

	unuseful_words.push_back("ein");
	unuseful_words.push_back("eine");
	unuseful_words.push_back("einer");
	unuseful_words.push_back("eines");
	unuseful_words.push_back("einem");

	unuseful_words.push_back("fr");
	unuseful_words.push_back("für");
	unuseful_words.push_back("fuer");

	for (unsigned int x = 0; x < unuseful_words.size(); x++) {
		jeliza::string unuseful_word(" " + unuseful_words[x] + " ");

		jeliza::string better_question(question.replace(unuseful_word, jeliza::string(" ")));

		if (better_question.get_strip().size() > 0) {
			question.set(better_question);
		}
	}

	return question.get_strip();
}

/*
 * Ist "ch" in "all"?
 */
bool is_in (string all, char ch) {
    std::string::iterator iter = find(all.begin(), all.end(), ch);

    if (iter == all.end()) {
        return false;
    }
    return true;
}

const std::string vokale("aeiouAEIOU");
const std::string konsonanten("bcdfghjklmnpqrstvwxyzBCDFGHJKLMNPQRSTVWXYZ");
const jeliza::string prefixe("ab an auf aus be bei dar ein er ent ge miss hin nach nieder ueber ber um unter ver vor weg zer zu");

/*
 * Entfernt Praefixe
 */
jeliza::string jeliza::spo_split::delete_prefixes(jeliza::string s1) {
    if (s1.size() < 6) {
        return s1;
    }

    jeliza::strings v_prefixe;
    jeliza::string prefixe_tmp(prefixe);
    util::split(prefixe_tmp, std::string(" "), v_prefixe);

    for (int x = 0; x < 3; x++) {
        for (unsigned int y = 0; y < v_prefixe.size(); y++) {
            if (s1.get_lower().startswith(v_prefixe[y])) {
                s1.replace_first_nocase(v_prefixe[y].get_lower(), jeliza::string(""));
            }
        }
    }

    return s1;
}

jeliza::spo_split& jeliza::spo_split::get() {
    return instance;
}

/*
 * Ermittelt den Wortstamm
 */
jeliza::string jeliza::spo_split::detect_word_base(jeliza::string s1) {
    jeliza::string stamm;

    jeliza::string orig_s1(s1);
    s1 = delete_prefixes(s1);

    int i = 0;
    if (is_in(konsonanten, s1[0])) {
        while (is_in(konsonanten, s1[i])) {
            stamm.ref() += s1[i];
            i++;
        }
    }
    while (is_in(vokale, s1[i])) {
        stamm.ref() += s1[i];
        i++;
    }
    while ((s1.size() - i > 2 || s1.size() < 4 || s1[i] == s1[i-1]) && is_in(konsonanten, s1[i])) {
        stamm.ref() += s1[i];
        i++;
    }

    return stamm;
}

/*
 * Gleicher Wortstamm?
 */
bool jeliza::spo_split::has_same_word_base(jeliza::string s1, jeliza::string s2, jeliza::string stamm1, jeliza::string stamm2) {
    if (s1 == s2) {
        return true;
    }

    if (s1 == stamm1) {
        return false;
    }
    if (s2 == stamm2) {
        return false;
    }

    if (stamm1 == stamm2) {
        return true;
    }

    s1.strip();
    s1.replace("_", " ");

    s2.strip();
    s2.replace("_", " ");

    if (s1.size() < 1 && s2.size() > 0) {
        return false;
    }
    if (s2.size() < 1 && s1.size() > 0) {
        return false;
    }

    if (s1 == s2) {
        return true;
    }

    bool valid_s1 = s1.get_lower() == s1;
    bool valid_s2 = s2.get_lower() == s2;
    if (valid_s1 != valid_s2) {
        return false;
    }

    if (s1.size() > 18) {
        return false;
    }
    if (s2.size() > 18) {
        return false;
    }

    return false;
}

/*
 * Sind sich die std::strings ziemlich aehnlich?
 */
bool jeliza::is_similar(jeliza::string s1, jeliza::string s2) {
    if (s1 == s2) {
        return true;
    }

    s1.lower();
    s1.strip();
    s1.replace("_", " ");

    s2.lower();
    s2.strip();
    s2.replace("_", " ");

    if (s1.size() < 1 && s2.size() > 0) {
        return false;
    }
    if (s2.size() < 1 && s1.size() > 0) {
        return false;
    }

    if (s1 == s2) {
        return true;
    }
    if (s1.contains(s2)) {
        return true;
    }
    if (s2.contains(s1)) {
        return true;
    }
    return false;
}


class number_format
{
    std::string What;

    public:
    number_format() : What("Typtransformation misslungen")
    {
    }

    number_format(std::string what) : What(what)
    {
    }

    std::string what(void)
    {
      return What;
    }
};

long string_to_long(std::string str)
{
  if (str.empty())
    throw number_format();
  char pos = 0;
  if (str[0] == '-')
    ++pos;
  if (str[pos] < 48 || str[pos] > 57)
    throw number_format();
  return std::strtol(str.c_str(), NULL, 10);
}

/*
 * Bestimmt die wordtype, wobei Pronomen und Adjektive als Nomen gewertet werden
 */
jeliza::wordtype jeliza::spo_split::detect_word_type_wrapper(jeliza::string word2, bool atBeginningOfSentence) {
    wordtype ret = detect_word_type(word2, atBeginningOfSentence);
    if (ret == PP) {
        ret = NOUN;
    }
    return ret;
}

jeliza::wordtype jeliza::detect_word_type_from_string(jeliza::string desc, bool atBeginningOfSentence) {
    desc.lower();
    if (desc.contains("nomen")) {
        cout << "- Downloaded word type: " << "noun" << endl;
        return jeliza::NOUN;
    }
    if (desc.contains("pronomen")) {
        cout << "- Downloaded word type: " << "noun" << endl;
        return jeliza::NOUN;
    }
    if (desc.contains("interjektion")) {
        cout << "- Downloaded word type: " << "interjection" << endl;
        return jeliza::INTER;
    }
    if (desc.contains("adv")) {
        cout << "- Downloaded word type: " << "adv" << endl;
        return jeliza::ADJ;
    }
    if (desc.contains("verb")) {
        cout << "- Downloaded word type: " << "verb" << endl;
        return jeliza::VERB;
    }
    if (desc.contains("adj")) {
        cout << "- Downloaded word type: " << "adj" << endl;
        return jeliza::ADJ;
    }
    if (desc.contains("position")) {
        cout << "- Downloaded word type: " << "prep" << endl;
        return jeliza::PREP;
    }
    return jeliza::NO_WORD_TYPE;
}

jeliza::genus jeliza::detect_genus_from_string(jeliza::string desc, bool atBeginningOfSentence) {
    desc.lower();
    if (desc.contains("neutr")) {
        cout << "- Downloaded genus: " << "neutrum" << endl;
        return jeliza::THING;
    }
    if (desc.contains("masku")) {
        cout << "- Downloaded genus: " << "maskulin" << endl;
        return jeliza::MALE;
    }
    if (desc.contains("feninin")) {
        cout << "- Downloaded genus: " << "feninin" << endl;
        return jeliza::FEMALE;
    }
    return jeliza::NO_GENUS;
}



/*
 * download word type from canoo.net
 */
jeliza::wordtype jeliza::download_word_type(jeliza::string word, bool atBeginningOfSentence) {
    if (offline_mode or word.contains("_")) {
        return jeliza::NO_WORD_TYPE;
    }
    std::string url("http://www.canoo.net/services/Controller?input=" + word.ref() + "&MenuId=Search&service=canooNet&lang=de");
    jeliza::msleep(2000);

    jeliza::string all(jeliza::download(url));
    all = all.replace("\r", "");
    jeliza::strings lines;
    util::split(all, "\n", lines);

    jeliza::strings not_correct_but_conjugated_last;
    jeliza::strings not_correct_but_conjugated;

    for (unsigned int x = 0; x < lines.size(); ++x) {
        jeliza::string line(lines[x]);
//        cout << line.ref() << endl;

        if (line.contains("<td class=\"Indent2\"><")) {
			continue;
        }
		if (!line.contains("<td class=\"Indent2\">")) {
			continue;
		}
		line = line.toASCII();
		cout << "Line: " << line.ref() << endl;
        line = line.replace("<td class=\"Indent2\">", "");
		cout << "Line: " << line.ref() << endl;


		jeliza::strings p;
		util::split(line, ":", p);
		if (p.size() < 2) continue;
		if (p[0].size() < 1) continue;
		if (word.size() < 1) continue;
		p[0].strip();
		cout << p[0].ref()
             << endl
             << p[0].replace_first(s("ge"), s("")).get_lower().ref()
             << endl
             << word.ref()
             << endl;

		if ((p[0].get_lower() == p[0] && word.get_lower() == word) || atBeginningOfSentence) {
            jeliza::wordtype wt = detect_word_type_from_string(line, atBeginningOfSentence);
		    if ((p[0].get_lower() == word.get_lower() || word.replace_first(s("ge"), s("")).get_lower() == p[0].get_lower()) && wt != ADJ) {
                if (wt != jeliza::NO_WORD_TYPE) {
                    cout << "- Downloaded word type: " << wt << endl;
                    return wt;
                }
		    }
		    else {
		        not_correct_but_conjugated.push_back(line);
		    }
		}
		jeliza::string p0_2(p[0]);
		jeliza::string word_3(word.replace_first_nocase(s("ge"), s("")));
		jeliza::string word_2(word);
		p0_2[0] = toupper(p0_2[0]);
		word_3[0] = toupper(word_3[0]);
		word_2[0] = toupper(word_2[0]);
		if ((p0_2 == p[0] && word_2 == word) || atBeginningOfSentence) {
            jeliza::wordtype wt = detect_word_type_from_string(line, atBeginningOfSentence);
		    if ((p0_2.get_lower() == word_2.get_lower() || word_3.get_lower() == p0_2.get_lower()) && wt != ADJ) {
                if (wt != jeliza::NO_WORD_TYPE) {
                    cout << "- Downloaded word type: " << wt << endl;
                    return wt;
                }
		    }
		    else {
		        not_correct_but_conjugated.push_back(line);
		    }
		}
        not_correct_but_conjugated_last.push_back(line);
    }
    for (unsigned int x = 0; x < not_correct_but_conjugated.size(); ++x) {
        jeliza::wordtype wt = detect_word_type_from_string(not_correct_but_conjugated[x], atBeginningOfSentence);
        if (wt != jeliza::NO_WORD_TYPE && wt == PREP) {
            cout << "- Downloaded word type (4): " << wt << endl;
            return wt;
        }
    }
    for (unsigned int x = 0; x < not_correct_but_conjugated.size(); ++x) {
        jeliza::wordtype wt = detect_word_type_from_string(not_correct_but_conjugated[x], atBeginningOfSentence);
        if (wt != jeliza::NO_WORD_TYPE && wt == VERB) {
            cout << "- Downloaded word type (2): " << wt << endl;
            return wt;
        }
    }
    for (unsigned int x = 0; x < not_correct_but_conjugated.size(); ++x) {
        jeliza::wordtype wt = detect_word_type_from_string(not_correct_but_conjugated[x], atBeginningOfSentence);
        if (wt != jeliza::NO_WORD_TYPE) {
            cout << "- Downloaded word type (3): " << wt << endl;
            return wt;
        }
    }
    for (unsigned int x = 0; x < not_correct_but_conjugated_last.size(); ++x) {
        jeliza::wordtype wt = detect_word_type_from_string(not_correct_but_conjugated_last[x], atBeginningOfSentence);
        if (wt != jeliza::NO_WORD_TYPE) {
            cout << "- Downloaded word type (last): " << wt << endl;
            return wt;
        }
    }

    return jeliza::NO_WORD_TYPE;
}

/*
 * download genus from canoo.net
 */
jeliza::genus jeliza::download_genus(jeliza::string word, bool atBeginningOfSentence) {
    if (offline_mode) {
        return jeliza::NO_GENUS;
    }
    std::string url("http://www.canoo.net/services/Controller?input=" + word.ref() + "&MenuId=Search&service=canooNet&lang=de");
    jeliza::msleep(2000);

    jeliza::string all(jeliza::download(url));
    all = all.replace("\r", "");
    jeliza::strings lines;
    util::split(all, "\n", lines);

    for (unsigned int x = 0; x < lines.size(); ++x) {
        jeliza::string line(lines[x]);
//        cout << line.ref() << endl;

        if (line.contains("<td class=\"Indent2\"><")) {
			continue;
        }
		if (!line.contains("<td class=\"Indent2\">")) {
			continue;
		}
		cout << "Line: " << line.ref() << endl;

        line = line.replace("<td class=\"Indent2\">", "");

		jeliza::strings p;
		util::split(line, ":", p);
		if (p.size() < 2) continue;
		if (p[0].size() < 1) continue;
		if (word.size() < 1) continue;
		p[0].strip();
		if ((p[0].get_lower() == p[0] && word.get_lower() == word) || atBeginningOfSentence) {
            jeliza::genus wt = detect_genus_from_string(line, atBeginningOfSentence);
            if (wt != jeliza::NO_GENUS) {
                cout << "- Downloaded word type: " << wt << endl;
                return wt;
            }
		}
		jeliza::string p0_2(p[0]);
		jeliza::string word_2(word);
		p0_2[0] = toupper(p0_2[0]);
		word_2[0] = toupper(word_2[0]);
		if ((p0_2 == p[0] && word_2 == word) || atBeginningOfSentence) {
            jeliza::genus wt = detect_genus_from_string(line, atBeginningOfSentence);
            if (wt != jeliza::NO_GENUS) {
                cout << "- Downloaded word type: " << wt << endl;
                return wt;
            }
		}


    }
    return jeliza::NO_GENUS;
}

/*
 * Bestimmt die wordtype
 */
jeliza::wordtype jeliza::spo_split::detect_word_type(jeliza::string word2, bool atBeginningOfSentence, bool noun_automatism) {
	word2 = word2.replace("{", "");
	word2 = word2.replace("}", "");
	word2 = word2.replace("/", "");
	word2 = word2.replace("|", "");
    if (word2.get_strip().size() == 0) {
        return NO_WORD_TYPE;
    }
    if (word2.get_lower().contains("http:")) {
        return NOUN;
    }
    if (word2.get_lower().contains("\"")) {
        return NOUN;
    }
    if (word2.get_lower().contains("'")) {
        return NOUN;
    }
    std::string word = word2.ref();
    if (word == "?" || word == "." || word == "," || word == ";") {
        return NO_WORD_TYPE;
    }

    word2 = word2.replace(".", "").replace("?", "").replace("!", "").replace(",", "").replace(";", "");
    word = word2.ref();

    std::string str;
    std::string word_low = word2.get_lower().ref();
    if (word_low == "und" || word_low == "oder" || word_low == "or" || word_low == "and") {
        return LINKING;
    }
    if (word_low == "keine" || word_low == "kein" || word_low == "keines" || word_low == "keiner" || word_low == "keinem" || word_low == "keinen") {
        return ADJ;
    }
    if (word_low == "__to__" || word_low == "_to_" || word_low == "wem" || word_low == "wer" || word_low == "wen" || word_low == "was" || word_low == "wie") {
        return QUESTIONWORD;
    }

    if (word_low == "anderen" || word_low == "denselben" || word_low == "dengleichen") {
        return ADJ;
    }

    if (word2.get_lower().contains("_")) {
        return NOUN;
    }
//    cout << "word2.ref(): " << word2.ref() << endl
//         << "word_low:    " << word_low << endl
//         << "beginning:   " << atBeginningOfSentence << endl;
    if (atBeginningOfSentence) {
        if (cache_noun_or_not[word_low] == NOUN) {
            return NOUN;
        }
        if (cache_noun_or_not[word_low] == VERB) {
            return VERB;
        }
        if (cache_noun_or_not[word_low] == ADJ) {
            return ADJ;
        }
        if (cache_noun_or_not[word_low] != WORDTYPE_UNKNOWN) {
            jeliza::wordtype wa1;
            jeliza::wordtype wa2;
            if (word2.ref() != word2.get_lower().ref()) {
                wa1 = detect_word_type(word2.get_lower(), false, false);
                jeliza::string word3(word2);
                word3[0] = toupper(word3[0]);
                wa2 = detect_word_type(word3, false, false);
                cout << word2.ref() << " != " << word3.ref() << endl
                     << wa1 << " != " << wa2 << endl;
                if (wa1 == QUESTIONWORD || wa2 == QUESTIONWORD) {
                    return QUESTIONWORD;
                }
            }
            if (wa1 != wa2 && word2.ref() != word2.get_lower().ref()) {
//                while (jeliza::next_noun_or_not != "-1" and jeliza::next_noun_or_not != "0") {
//                    jeliza::msleep(500);
//                    cout << "jeliza::next_noun_or_not:" << jeliza::next_noun_or_not << endl;
//                }

                jeliza::next_noun_or_not = "0";

                unsigned int a = 5;
                while (jeliza::next_noun_or_not[0] == '0') {
                    jeliza::comm_send("GET_NOUN_OR_NOT:" + word2.ref());
                    cout << "GET_NOUN_OR_NOT:" + word2.ref() << endl;

                    ++a;
                    jeliza::msleep(500);
                    if (jeliza::fast_exit) {
                        jeliza::next_noun_or_not = "-1";
                        return NO_WORD_TYPE;
                    }
                }
                jeliza::comm_send("CLEAN_NOUN_OR_NOT:.");
                cout << "500" << endl;

                if (jeliza::next_noun_or_not.size() > 0) {
                    jeliza::force_noun_or_not = word2.ref();
                    if (jeliza::next_noun_or_not[0] == '1') {
                        jeliza::next_noun_or_not = "-1";
                        cache_noun_or_not[word_low] = NOUN;
                        return NOUN;
                    }
                    if (jeliza::next_noun_or_not[0] == '2') {
                        jeliza::next_noun_or_not = "-1";
                        cache_noun_or_not[word_low] = wa1;
                        return wa1;
                    }
                    cache_noun_or_not[word_low] = WORDTYPE_UNKNOWN;
                }
                else {
                    cout << "Bug gefunden! Nichts eingegeben in GUI oder falsch gesendet!" << endl;
                }
                cout << "501" << endl;

                jeliza::next_noun_or_not = "-1";
                cout << "502" << endl;
            }
        }
    }

    std::string word_substrd = word.substr(0, word.size() - 3);

    jeliza::string prefixe_entfernt_word = delete_prefixes(word2);
    jeliza::string prefixe_entfernt_word_low = delete_prefixes(word2.get_lower());

    ifstream iVerbs(("lang_" + jeliza::get_lang_str() + "/wsh/verbs.wsh").ref().c_str());
    vector<std::string> vVerbs;
    while (iVerbs) {
        iVerbs >> str;
        vVerbs.push_back(str);
    }

    if (find(vVerbs.begin(), vVerbs.end(), word) != vVerbs.end()) {
        return VERB;
    }

    // Artikel
    ifstream iArts(("lang_" + jeliza::get_lang_str() + "/wsh/arts.wsh").ref().c_str());
    vector<std::string> vArts;
    while (iArts) {
        iArts >> str;
        vArts.push_back(str);
    }

    if (find(vArts.begin(), vArts.end(), word_low) != vArts.end()) {
        return ART;
    }

    // Pronomen
    ifstream iPP(("lang_" + jeliza::get_lang_str() + "/wsh/pp.wsh").ref().c_str());
    vector<std::string> vPP;
    while (iPP) {
        iPP >> str;
        vPP.push_back(str);
    }

    if (find(vPP.begin(), vPP.end(), word_low) != vPP.end()) {
        return PP;
    }

    // Praepositions
    ifstream iPreps(("lang_" + jeliza::get_lang_str() + "/wsh/preps.wsh").ref().c_str());
    vector<std::string> vPreps;
    while (iPreps) {
        iPreps >> str;
        vPreps.push_back(str);
    }

    if (find(vPreps.begin(), vPreps.end(), word_low) != vPreps.end()) {
        return PREP;
    }

    jeliza::string vec_0;
    if (!atBeginningOfSentence) {
        if (jeliza::findit(&wordtype_equal, word2.ref()).size() > 0) {
            vec_0.set(jeliza::findit(&wordtype_equal, word2.ref())[0]);
        }
//        else if (jeliza::findit(&wordtype_ohne_prefix, prefixe_entfernt_word.ref()).size() > 0) {
//            vec_0.set(jeliza::findit(&wordtype_ohne_prefix, prefixe_entfernt_word.ref())[0]);
//        }
    }
    else {
        if (jeliza::findit(&wordtype_equal_low, word_low).size() > 0) {
            vec_0.set(jeliza::findit(&wordtype_equal_low, word_low)[0]);
        }
        else if (jeliza::findit(&wordtype_equal, word2.ref()).size() > 0) {
            vec_0.set(jeliza::findit(&wordtype_equal, word2.ref())[0]);
        }
//        else if (jeliza::findit(&wordtype_ohne_prefix, prefixe_entfernt_word.ref()).size() > 0) {
//            vec_0.set(jeliza::findit(&wordtype_ohne_prefix, prefixe_entfernt_word.ref())[0]);
//        }
    }
    if (vec_0.size() > 0) {
        if (vec_0 == "prep") {
            return PREP;
        }
        if (vec_0 == "fw") {
            return QUESTIONWORD;
        }

        if (vec_0 == "verb") {
            return VERB;
        }
        if (vec_0 == "vt") {
            return VERB;
        }
        if (vec_0 == "vi") {
            return VERB;
        }
        if (vec_0 == "n") {
            return NOUN;
        }
        if (vec_0 == "f,m") {
            return NOUN;
        }
        if (vec_0 == "f,n") {
            return NOUN;
        }
        if (vec_0 == "f,n,m") {
            return NOUN;
        }
        if (vec_0 == "f,m,n") {
            return NOUN;
        }
        if (vec_0 == "m,f") {
            return NOUN;
        }
        if (vec_0 == "m,n") {
            return NOUN;
        }
        if (vec_0 == "m,f,n") {
            return NOUN;
        }
        if (vec_0 == "m,n,f") {
            return NOUN;
        }
        if (vec_0 == "m,") {
            return NOUN;
        }
        if (vec_0 == "f,") {
            return NOUN;
        }
        if (vec_0 == "n,") {
            return NOUN;
        }
        if (vec_0 == "inter") {
            return INTER;
        }
        if (vec_0 == "n,pl") {
            return NOUN;
        }

        if (vec_0 == "adj") {
            return ADJ;
        }
        if (vec_0 == "adv") {
            return ADJ;
        }
        if (vec_0 == "pron") {
            return PP;
        }
        if (vec_0 == "ppron") {
            return NOUN;
        }
        if (vec_0 == "nothing") {
            return UNIMPORTANT;
        }
    }

    try {
        string_to_long(word_low);
        return ADJ;
    }
    catch (number_format &ex) {
//        std::cerr << ex.what() << std::endl;
    }

    if (!atBeginningOfSentence && word_low != word2.ref() && noun_automatism) {
        return NOUN;
    }

    jeliza::wordtype downloaded = download_word_type(word2, atBeginningOfSentence);
    if (downloaded != NO_WORD_TYPE) {
        jeliza::force_answer_word = word2.ref();
        if (downloaded == jeliza::NOUN) jeliza_impl_instance.add_word('2');
//        if (downloaded == jeliza::PRONOUN) jeliza_impl_instance.add_word('2');
        if (downloaded == jeliza::VERB) jeliza_impl_instance.add_word('1');
        if (downloaded == jeliza::ADJ) jeliza_impl_instance.add_word('3');
        if (downloaded == jeliza::PREP) jeliza_impl_instance.add_word('6');

        cout << "Downloaded: " << downloaded << endl;

        return downloaded;
    }
    cout << "Downloaded == NO_WORD_TYPE " << endl;

    while (not(jeliza::next_word_type == "-1") || jeliza::next_word_type == "0") {
            jeliza::msleep(500);
    }

    {
        jeliza::next_word_type = "0";

        unsigned int a = 5;
        while (jeliza::next_word_type[0] == '0') {
            jeliza::comm_send("GET_WORD_TYPE:" + word2.ref());
            cout << "GET_WORD_TYPE:" + word2.ref() << endl;

            ++a;
            jeliza::msleep(500);
            if (jeliza::fast_exit) {
            	return NO_WORD_TYPE;
            }
        }
        jeliza::comm_send("CLEAN_WORD_TYPE:.");
        cout << "500" << endl;

        if (jeliza::next_word_type.size() > 0) {
            jeliza::force_answer_word = word2.ref();
            jeliza_impl_instance.add_word(jeliza::next_word_type[0]);
        }
        else {
            cout << "Bug gefunden! Nichts eingegeben in GUI oder falsch gesendet!" << endl;
        }
        cout << "501" << endl;

        jeliza::next_word_type = "-1";
        cout << "502" << endl;
        return detect_word_type(word2, atBeginningOfSentence);
    }


    return NO_WORD_TYPE;
}

jeliza::genus jeliza::spo_split::define_genus(jeliza::string word2) {
	word2 = word2.replace("{", "");
	word2 = word2.replace("}", "");
    if (word2.get_strip().size() == 0) {
        return NO_GENUS;
    }
    std::string word = word2.ref();
    if (word == "?" || word == "." || word == "," || word == ";") {
        return NO_GENUS;
    }

    word2 = word2.replace(".", "").replace("?", "").replace("!", "").replace(",", "").replace(";", "").replace("\"", "").replace("'", "");
    word = word2.ref();

    std::string str;
    std::string word_low = word2.get_lower().ref();
    if (word_low == "und" || word_low == "oder" || word_low == "or" || word_low == "and") {
        return NO_GENUS;
    }


    std::string genus = (*jeliza::spo_split::get().genus_index)[word];
    if (genus.size()) {
    	if (genus == "male") cout << word << " MALE" << endl;
    	if (genus == "female") cout << word << " FEMALE" << endl;
    	if (genus == "thing") cout << word << " THING" << endl;
    	if (genus == "male") return MALE;
    	if (genus == "female") return FEMALE;
    	if (genus == "thing") return THING;
    	cout << "Unbekannter genus: " << genus << endl;
    }
	cout << "Unbekannter genus: " << genus << endl;


    jeliza::genus downloaded = download_genus(word2, false);
    if (downloaded != NO_GENUS) {
        if (downloaded == jeliza::MALE) jeliza_impl_instance.add_genus('1', word);
        if (downloaded == jeliza::FEMALE) jeliza_impl_instance.add_genus('2', word);
        if (downloaded == jeliza::THING) jeliza_impl_instance.add_genus('3', word);

        cout << "Downloaded: " << downloaded << endl;

        return downloaded;
    }
    cout << "Downloaded == NO_WORD_TYPE " << endl;



    while (not(jeliza::next_genus == "-1") || jeliza::next_genus == "0") {
            jeliza::msleep(500);
    }

    {
        jeliza::next_genus = "0";

        unsigned int a = 5;
        while (jeliza::next_genus[0] == '0') {
            jeliza::comm_send("GET_GENUS:" + word2.ref());
            cout << "GET_GENUS:" + word2.ref() << endl;

            ++a;
            jeliza::msleep(500);
            if (jeliza::fast_exit) {
            	return NO_GENUS;
            }
        }
        jeliza::comm_send("CLEAN_GENUS:.");
        cout << "500" << endl;

        if (jeliza::next_genus.size() > 0) {
            jeliza_impl_instance.add_genus(jeliza::next_genus[0], word);
        }
        else {
            cout << "Bug gefunden! Nichts eingegeben in GUI oder falsch gesendet!" << endl;
        }
        cout << "501" << endl;

        jeliza::next_genus = "-1";
        cout << "502" << endl;
        return define_genus(word2);
    }


    return NO_GENUS;
}



jeliza::worddef* next_word_if(vector<jeliza::worddef>& woerter, jeliza::wordtype soll_wa, unsigned int index) {
    for (unsigned int x = index; x < woerter.size(); x++) {
        if (woerter[x].second == soll_wa) {
            return (&woerter[x]);
        }
    }
    return NULL;
}

unsigned int next_word_index_if(vector<jeliza::worddef>& woerter, jeliza::wordtype soll_wa, unsigned int index) {
    for (unsigned int x = index; x < woerter.size(); x++) {
        if (woerter[x].second == soll_wa) {
            return x;
        }
    }
    return -1;
}

bool pred_lexicographical_compare (const jeliza::string& lhs, const jeliza::string& rhs) {
    return std::lexicographical_compare(lhs.get_lower().begin(), lhs.size()>10 ?
            lhs.get_lower().begin()+10 : lhs.get_lower().end(), rhs.get_lower().begin(), rhs.size()>10 ? rhs.get_lower().begin()+10 :
            rhs.get_lower().end());
}

/*
 * Satzauftrennungsdatenbank initialisieren
 */
void jeliza::spo_split::init_sentences_splitted() {
    {
        sentences_splitted = map<vector<jeliza::wordtype>, vector<std::pair<jeliza::wordtype, std::pair<jeliza::expression_part, jeliza::sentence_part> > > >();

        ifstream i("temp/sentences.tmp");
        jeliza::string line;
        unsigned long a = 0;
        vector<jeliza::string> parts;
        while (i) {
            getline(i, line.ref());

            ++a;
            if (a % 100 == 0) {
                jeliza::update_prozent(100.0 / 80000.0 * a);
            }

            vector<vector<std::pair<wordtype, std::pair<expression_part, sentence_part> > > > musterLines;
            vector<vector<wordtype> > keys;

            vector<std::pair<wordtype, std::pair<expression_part, sentence_part> > > musterLine;
            vector<wordtype> key;
            musterLines.push_back(musterLine);
            keys.push_back(key);

            util::split(line, std::string(" "), parts);
            wordtype i = NO_WORD_TYPE;
            expression_part s = NO_CLAUSE;
            sentence_part os = NO_SENTENCE;
            for (unsigned int x = 0; x < parts.size(); ++x) {
                parts[x].strip();
                if (parts[x].contains("'''")) {
                    if (parts[x].contains("SUBJ")) {
                        s = SUBJ;
                    }
                    if (parts[x].contains("OBJ")) {
                        s = OBJ;
                    }
                    if (parts[x].contains("PRE")) {
                        s = PRAE;
                    }
                    if (parts[x].contains("ZUS")) {
                        s = PROPERTIES;
                    }

                    if (parts[x].contains("HS")) {
                        os = MAIN_CLAUSE;
                    }
                    if (parts[x].contains("NS")) {
                        os = SUB_CLAUSE;
                    }
                    if (parts[x].contains("FS")) {
                        os = MAIN_CLAUSE;
                    }

                    continue;
                }

                if (parts[x].contains("ART")) i = ART;
                else if (parts[x].contains("ADJ")) i = ADJ;
                else if (parts[x].contains("N")) i = NOUN;
                else if (parts[x].contains("SPO")) i = SPO;
                else if (parts[x].contains("VERB")) i = VERB;
                else if (parts[x].contains("QW")) i = QUESTIONWORD;
                else if (parts[x].contains("PRPO")) i = PREP;
                else if (parts[x].contains("KOMMA")) i = KOMMA;
                else if (parts[x].contains("LINKING")) i = LINKING;

                for (unsigned int y = 0; y < musterLines.size(); ++y) {
                    musterLines[y].push_back(make_pair(i, make_pair(s, os)));
                    keys[y].push_back(i);
                }
            }

            for (unsigned int x = 0; x < musterLines.size(); ++x) {
                sentences_splitted[keys[x]] = musterLines[x];
            }
        }
    }
}

bool is_first_name (std::string s) {
    ifstream iii(("lang_" + jeliza::get_lang_str() + "/wsh/first_names.wsh").ref().c_str());
    while (iii) {
        jeliza::string fn;
        iii >> fn.ref();
        fn.strip();
//        cout << "Vorname: '" << fn.ref() << "' != '" << s << "'" << endl;
        if (fn.get_lower().ref() == jeliza::string(s).get_lower().get_strip().ref()) {
            iii.close();
            return true;
        }
    }
    iii.close();
    return false;
}

/*
 * Trennt einen Satz auf
 */
std::pair<vector<jeliza::db_sentence>, vector<jeliza::db_sentence> > jeliza::spo_split::split_spo (jeliza::string s) {
	s = s.replace("{", "");
	s = s.replace("}", "");
	s = s.replace("  ", " ");
	s = s.replace("  ", " ");
	s = s.replace("  ", " ");
	s = s.replace("  ", " ");

	jeliza::string _s(s);
	bool in_quotation = false;
	for (unsigned int x = 0; x < _s.size(); ++x) {
	    if (in_quotation == false && (_s[x] == '"' or _s[x] == '\'')) {
	        in_quotation = true;
	        s[x] = '"';
	        continue;
	    }
	    if (in_quotation && (_s[x] == ' ')) {
	        s[x] = '_';
	    }
	    if (in_quotation == true && (_s[x] == '"' or _s[x] == '\'')) {
	        in_quotation = false;
	        s[x] = '"';
	        continue;
	    }
	}

    jeliza::strings words;
    /*
     * Diese Zeichen werden entfernt
     */
    std::string findd = ".()/-:;!?\n\r ";
    /*
     * Das Komma wird zu einem ganzen Wort
     */
    s = s.replace_nocase_safe(jeliza::string(", "), jeliza::string(","));
    s = s.replace_nocase_safe(jeliza::string(" ,"), jeliza::string(","));
    s = s.replace_nocase_safe(jeliza::string(","), jeliza::string(" , "));
    s = s.replace("  ", " ");
    /*
     * Zeichen von oben entfernen
     */
    util::split(s, findd, words);
    util::join(s, std::string(" "), words);

    {
        /*
         * Adverbiale Bestimmungen und alles andere dividen
         */

        jeliza::strings words;
        util::split(s, std::string(" "), words);
        vector<std::pair<worddef, jeliza::string> > woerter;
        jeliza::string adverbiale_bestimmungen;
        jeliza::string subj_hinten_dran;
        bool adverbiale_bestimmung = false;
        jeliza::string lastWord;
        wordtype lastWa = WORDTYPE_UNKNOWN;
        wordtype firstWa = WORDTYPE_UNKNOWN;
        jeliza::string advb_bestim;
        for (unsigned int x = 0; x < words.size(); ++x) {
            jeliza::string word(words[x]);
            jeliza::string word_next;
            unsigned int d;
            for (d = x; d < words.size(); ++d) {
                jeliza::string word(words[d]);
                wordtype wa = (word == ",") ? KOMMA : detect_word_type_wrapper(word, d == 0);
                if (wa == PP) wa = NOUN;
                if (wa == NOUN) {
                	word_next.set(words[d]);
                	break;
                }
            }
            jeliza::string word_tmp(words[x]);
            word.strip();
            word_tmp.strip();
            wordtype wa = (word == ",") ? KOMMA : detect_word_type_wrapper(word, x == 0);
            wordtype wa_echt = (word == ",") ? KOMMA : detect_word_type(word, x == 0);

            if (is_first_name(word.ref()) && x+1 < words.size()) {
                words[x+1] = word + "_" + words[x+1];
                continue;
            }

            int adverbiale_bestimmung_backup = 2;
            if (is_first_name(word.get_lower().ref()) || is_first_name(lastWord.get_lower().ref())) {
                adverbiale_bestimmung_backup = adverbiale_bestimmung;
            }


            if (wa_echt == UNIMPORTANT) {
                continue;
            }

            if (firstWa == WORDTYPE_UNKNOWN) {
                firstWa = wa;
            }
            if (wa == PP) wa = NOUN;

            if (wa == ART) {
                word_tmp += " - Artikel";
            }
            if (wa == ADJ) {
                word_tmp += " - Adjektiv";
            }
            if (wa == NOUN) {
                word_tmp += " - NOUN";
            }
            if (wa == VERB) {
                word_tmp += " - Verb";
            }
            if (wa == QUESTIONWORD) {
                word_tmp += " - questionwort";
            }
            if (wa == PREP) {
                word_tmp += " - Praeposition";
            }
            if (wa == SPO) {
                word_tmp += " - Subjekt/Verb/Objekt";
            }
            if (wa == KOMMA) {
                word_tmp += " - ein Komma";
            }


            if (wa == lastWa && wa == NOUN && adverbiale_bestimmung) {
                if (!is_first_name(word.get_lower().ref()) || is_first_name(lastWord.get_lower().ref())) {
                    adverbiale_bestimmung = false;
                }
            }

            if (wa == PREP) {
                adverbiale_bestimmung = true;
            }
            if (wa == VERB) {
                if (!is_first_name(word.get_lower().ref()) || is_first_name(lastWord.get_lower().ref())) {
                    adverbiale_bestimmung = false;
                }
            }

            if (wa == ART && jeliza::get_lang() == GERMAN) {
            	cout << "word_next:" << word_next.ref() << endl;
            	genus g = define_genus(word_next);

            	if (g == MALE) {
            		if (word.get_lower().ref() != "der"
            			&& word.get_lower().ref() != "den"
            			&& word.get_lower().ref() != "ein"
                		&& word.get_lower().ref() != "einen"
                        && word.get_lower().ref() != "mein"
                        && word.get_lower().ref() != "meines"
                        && word.get_lower().ref() != "dein"
                        && word.get_lower().ref() != "deines"
                        && word.get_lower().ref() != "sein"
                        && word.get_lower().ref() != "seines"
                        && word.get_lower().ref() != "unser"
                        && word.get_lower().ref() != "unseres"
                        && word.get_lower().ref() != "unsers"
                        && word.get_lower().ref() != "euer"
                        && word.get_lower().ref() != "eures"
                        && word.get_lower().ref() != "eueres"
                        && word.get_lower().ref() != "euers"
                        && word.get_lower().ref() != "ihr"
                        && word.get_lower().ref() != "ihres"
            		) {
                    	adverbiale_bestimmung = true;
            		}
            	}
            	if (g == FEMALE) {
            		if (word.get_lower().ref() != "die"
            			&& word.get_lower().ref() != "eine"
            			&& word.get_lower().ref() != "einer"
                            && word.get_lower().ref() != "meine"
                            && word.get_lower().ref() != "meiner"
                            && word.get_lower().ref() != "deine"
                            && word.get_lower().ref() != "deiner"
                            && word.get_lower().ref() != "seine"
                            && word.get_lower().ref() != "seiner"
                            && word.get_lower().ref() != "unsere"
                            && word.get_lower().ref() != "unserer"
                            && word.get_lower().ref() != "unserer"
                            && word.get_lower().ref() != "euere"
                            && word.get_lower().ref() != "eurer"
                            && word.get_lower().ref() != "euerer"
                            && word.get_lower().ref() != "euere"
                            && word.get_lower().ref() != "ihre"
                            && word.get_lower().ref() != "ihrer"
            		) {
                    	adverbiale_bestimmung = true;
            		}
            	}
            	if (g == THING) {
            		if (word.get_lower().ref() != "das"
            			&& word.get_lower().ref() != "eines"
            			&& word.get_lower().ref() != "ein"
                            && word.get_lower().ref() != "mein"
                            && word.get_lower().ref() != "meines"
                            && word.get_lower().ref() != "dein"
                            && word.get_lower().ref() != "deines"
                            && word.get_lower().ref() != "sein"
                            && word.get_lower().ref() != "seines"
                            && word.get_lower().ref() != "unser"
                            && word.get_lower().ref() != "unseres"
                            && word.get_lower().ref() != "unsers"
                            && word.get_lower().ref() != "euer"
                            && word.get_lower().ref() != "eures"
                            && word.get_lower().ref() != "eueres"
                            && word.get_lower().ref() != "euers"
                            && word.get_lower().ref() != "ihr"
                            && word.get_lower().ref() != "ihres"
            		) {
                    	adverbiale_bestimmung = true;
            		}
            	}
            }

            bool ist_einzelnes_adjektiv = false;
            if (wa_echt == ADJ) {
                bool hat_kein_nomen_dahinter = true;
                if (x + 1 >= words.size()) {
                    hat_kein_nomen_dahinter = true;
                }
                if (x + 1 < words.size()) {
                    wordtype wa__ = detect_word_type(words[x+1], false);
                    if (wa__ == NOUN || wa__ == PP) {
                        hat_kein_nomen_dahinter = false;
                    }
                }
                if (x + 2 < words.size()) {
                    wordtype wa__ = detect_word_type(words[x+2], false);
                    if (wa__ == NOUN || wa__ == PP) {
                        hat_kein_nomen_dahinter = false;
                    }
                }
                if (x + 3 < words.size()) {
                    wordtype wa__ = detect_word_type(words[x+3], false);
                    if (wa__ == NOUN || wa__ == PP) {
                        hat_kein_nomen_dahinter = false;
                    }
                }
                if (hat_kein_nomen_dahinter) {
                    ist_einzelnes_adjektiv = true;
                }
            }
            bool vor_questionwort = false;
            if (x + 1 < words.size()) {
                wordtype wa__ = detect_word_type(words[x+1], false);
                if (wa__ == QUESTIONWORD) {
                    vor_questionwort = true;
                    if (!is_first_name(word.get_lower().ref()) || is_first_name(lastWord.get_lower().ref())) {
                    adverbiale_bestimmung = false;
                }
                }
            }
            bool nach_questionwort = false;
            if (x - 1 < words.size() && (int)(x) - 1 >= 0) {
                wordtype wa__ = detect_word_type(words[x-1], false);
                if (
                        wa__ == QUESTIONWORD
                		&& (
                				not(words[x-1].get_strip().get_lower() == "wenn")
                				and not(words[x-1].get_strip().get_lower() == "falls")
                				and not(words[x-1].get_strip().get_lower() == "sobald")
                				and not(words[x-1].get_strip().get_lower() == "if")
                				and not(words[x-1].get_strip().get_lower() == "when")
                				and not(words[x-1].get_strip().get_lower() == "why")
                				and not(words[x-1].get_strip().get_lower() == "warum")
                				and not(words[x-1].get_strip().get_lower() == "wieso")
                				and not(words[x-1].get_strip().get_lower() == "weshalb")
                				and not(words[x-1].get_strip().get_lower() == "because")
                				and not(words[x-1].get_strip().get_lower() == "weil")
                		)) {
                    nach_questionwort = true;
                    if (!is_first_name(word.get_lower().ref()) || is_first_name(lastWord.get_lower().ref())) {
                    adverbiale_bestimmung = false;
                }
                }
            }
            bool verb_in_sicht = false;
            if (x < words.size()) {
                wordtype wa__ = detect_word_type(words[x], false);
                if (wa__ == VERB) {
                    verb_in_sicht = true;
                }
            }
            if (x + 1 < words.size()) {
                wordtype wa__ = detect_word_type(words[x+1], false);
                if (wa__ == VERB) {
                    verb_in_sicht = true;
                }
            }
            if (x + 2 < words.size()) {
                wordtype wa__ = detect_word_type(words[x+2], false);
                if (wa__ == VERB) {
                    verb_in_sicht = true;
                }
            }
            if (x + 3 < words.size()) {
                wordtype wa__ = detect_word_type(words[x+3], false);
                if (wa__ == VERB) {
                    verb_in_sicht = true;
                }
            }
/*
 *             if (x+3 >= words.size()) {
 * 		          verb_in_sicht = false;
 *             }
 */

            if (adverbiale_bestimmung_backup != 2) {
                adverbiale_bestimmung = adverbiale_bestimmung_backup;
            }


            if ((vor_questionwort || nach_questionwort) && (wa == ART || wa == NOUN || wa == ADJ || wa == PREP) && verb_in_sicht && ((( (int)(x)-1 >= 0) ? (words[x-1]) : jeliza::string()) != jeliza::string("dass"))) {
                advb_bestim += " ";
                advb_bestim += word;
                advb_bestim.strip();
            }
            else if (adverbiale_bestimmung) {
            	advb_bestim += " ";
                advb_bestim += word;
                advb_bestim.strip();
            }
            else if (ist_einzelnes_adjektiv) {
                advb_bestim += " ";
                advb_bestim += word;
                advb_bestim.strip();
            }
            else {
                if (woerter.size()) {
                	woerter[woerter.size() - 1].second += advb_bestim;
                    woerter.push_back(std::pair<worddef, jeliza::string>(worddef(word, wa), jeliza::string()));
                }
                else {
                    woerter.push_back(std::pair<worddef, jeliza::string>(worddef(word, wa), advb_bestim));
                }
                if (advb_bestim.size() != 0) {
                    advb_bestim.set("");
                }
                if (!is_first_name(word.get_lower().ref()) || is_first_name(lastWord.get_lower().ref())) {
                    if (!is_first_name(word.get_lower().ref()) || is_first_name(lastWord.get_lower().ref())) {
                    adverbiale_bestimmung = false;
                }
                }
            }

            if (wa == NOUN) {
                if (!is_first_name(word.get_lower().ref()) || is_first_name(lastWord.get_lower().ref())) {
                    if (!is_first_name(lastWord.get_lower().ref())) {
                        adverbiale_bestimmung = false;
                    }
                }
            }

            cout << adverbiale_bestimmung << ist_einzelnes_adjektiv << vor_questionwort << nach_questionwort << verb_in_sicht << is_first_name(lastWord.get_lower().ref())
                 << " " << word_tmp.ref() << wa << lastWord.ref() << endl;

            lastWa = wa;
            lastWord.set(word);
        }

        if (advb_bestim.size() != 0) {
            if (woerter.size() > 0) {
                woerter[woerter.size() - 1].second += " ";
                woerter[woerter.size() - 1].second += advb_bestim;
                woerter[woerter.size() - 1].second.strip();
                advb_bestim.set("");
            }
        }
        else {
            woerter.push_back(std::pair<worddef, jeliza::string>(worddef(jeliza::string(), lastWa), advb_bestim));
        }


        /*
         * Ueberpruefen, ob die Trenndatenbank zum divide verwendet werden kann
         *
         * Vorbereitung:
         */

//        log("Pruefen...");
        vector<std::pair<wordtype, std::pair<jeliza::expression_part, jeliza::sentence_part> > > musterLine;
        bool passt = false;

        vector<wordtype> key;
        map<unsigned int, unsigned int> index_lang_zu_kurz;
        for (unsigned int x = 0; x < woerter.size(); ++x) {
            wordtype wa = woerter[x].first.second;
            if (wa == ART) wa = SPO;
            if (wa == NOUN) wa = SPO;
            if (wa == VERB) wa = SPO;
            if (wa == ADJ) wa = SPO;
            if (wa == LINKING) wa = SPO;
            key.push_back(wa);
            index_lang_zu_kurz[x] = x;

            jeliza::string word_tmp;
            if (wa == PP) wa = NOUN;

            if (wa == ART) {
                word_tmp += " - article";
            }
            if (wa == ADJ) {
                word_tmp += " - adjective";
            }
            if (wa == NOUN) {
                word_tmp += " - noun";
            }
            if (wa == VERB) {
                word_tmp += " - verb";
            }
            if (wa == QUESTIONWORD) {
                word_tmp += " - question word";
            }
            if (wa == PREP) {
                word_tmp += " - preposition";
            }
            if (wa == SPO) {
                word_tmp += " - subject/verb/object";
            }
            if (wa == KOMMA) {
                word_tmp += " - komma";
            }
//            log(word_tmp.ref());
        }

        /*
         * Passt es?
         */
        /**
        musterLine = jeliza::spo_split::get().sentences_splitted[key];
        if (musterLine.size() > 0) {
            log("PASST!");
            passt = true;
        }
        **/

        /*
         * Alternativer und primitiver Aufdivider
         */
        /**
        if (!passt) {
            log("PASST NICHT!");

            vector<vector<jeliza::db_sentence> > ret;

            return ret;
        }
        **/

        std::pair<vector<jeliza::db_sentence>, vector<jeliza::db_sentence> > ret_list;
        ret_list.first = vector<jeliza::db_sentence>();
        ret_list.second = vector<jeliza::db_sentence>();

        /*
         * Datenbank konnte verwendet werden
         */
        jeliza::string subj;
        jeliza::string prae;
        jeliza::string obj;
        jeliza::string additional_things;
        jeliza::string questionwort;
        jeliza::string adv_bestm;
        unsigned int hatte_schon_subjekt = 0; // 0 == Subjekt, 1 == Objekt
//        cout << endl << endl << endl << endl;
        for (unsigned int x = 0; x < woerter.size(); ++x) {
            jeliza::string w(woerter[x].first.first);
            wordtype wa1 = woerter[x].first.second;

            /*
             * Nur den relevanten Satteil betrachten
             */
            /**
            cout << musterLine[index_lang_zu_kurz[x]].second.second << musterLine[x].second.second << w.ref() << endl;
            if (musterLine[index_lang_zu_kurz[x]].second.second != sentence_part) {
                continue;
            }
            if (wa1 == QUESTIONWORD && sentence_part != SUB_CLAUSE) {
                break;
            }
            **/

            if (wa1 == KOMMA) {
                subj.strip();
                prae.strip();
                prae.lower();
                obj.strip();
                questionwort.strip();
                additional_things.strip();
                subj_hinten_dran.strip();

                /*
                 * Verben: einheitliche Reihenfolge
                 */
                jeliza::strings verben;
                util::split(prae, std::string(" "), verben);
                sort (verben.begin(), verben.end(), pred_lexicographical_compare);
                util::join(prae, std::string(" "), verben);

                jeliza::db_sentence ret;

                ret.subject = subj;
                ret.verb = prae;
                ret.object = obj + " " + subj_hinten_dran;
                ret.object.strip();
                if (!ret.object.size()) {
                    ret.object.set("nothing {}");
                }
                if (!ret.subject.size()) {
                    ret.subject.set("nothing {}");
                }
                ret.quesword = questionwort.get_lower();
                ret.additional = additional_things + adverbiale_bestimmungen;

                subj.set("");
                prae.set("");
                obj.set("");
                questionwort.set("");
                additional_things.set("");
                subj_hinten_dran.set("");
                hatte_schon_subjekt = 0;

                if (ret.quesword.size()) {
                    ret_list.second.push_back(ret);
                }
                else {
                    ret_list.first.push_back(ret);
                }
            }

            /*
             * Pronomen sind hier Nomen
             */
            if (wa1 == PP) wa1 = NOUN;

            jeliza::strings words;
            util::split(w, std::string(" "), words);
            for (unsigned int y = 0; y < words.size(); ++y) {
                jeliza::string word(words[y]);

                if (word == ",") {
                    continue;
                }

                if (wa1 == PREP) {
                    additional_things += word + " ";
                	adverbiale_bestimmungen += " ";
                	adverbiale_bestimmungen += woerter[x].second;
                	adverbiale_bestimmungen.strip();
                }
                if (wa1 == QUESTIONWORD) {
                    questionwort += word;
                    questionwort += " ";
                    hatte_schon_subjekt = 0;
                    additional_things += subj + " " + obj + " ";
                    subj.set("");
                    obj.set("");
                	adverbiale_bestimmungen += " ";
                	adverbiale_bestimmungen += woerter[x].second;
                	adverbiale_bestimmungen.strip();
                }
                else if (wa1 == VERB) {
                    prae += word;
                    prae += " ";

                    subj.strip();
                    if (subj.size() > 0) {
                        hatte_schon_subjekt++;
                    }
                	adverbiale_bestimmungen += " ";
                	adverbiale_bestimmungen += woerter[x].second;
                	adverbiale_bestimmungen.strip();
                }
/*
                else if (musterLine[x].second.first == PROPERTIES) {
                    additional_things += word + " ";
                }
*/
                else {
                    wordtype wa_ = detect_word_type(word, x == 0);
                    wordtype wa_next;
                    if (y+1 < words.size()) {
                    	wa_next = detect_word_type(words[y+1], x == 0);
                    } else if (x+1 < woerter.size()) {
                    	jeliza::string w2(woerter[x+1].first.first);
                    	jeliza::strings words_2;
                        util::split(w2, std::string(" "), words_2);
                        if (words_2.size()) {
                        	wa_next = detect_word_type(words_2[0], x+1 == 0);
                        }
                    }

                    if (wa_ == LINKING) {
                        hatte_schon_subjekt--;
                    }
                    else if (is_first_name(word.get_lower().ref())) {
                        hatte_schon_subjekt--;
                    }

                    adv_bestm += " ";
                    adv_bestm += woerter[x].second;
                    adv_bestm.strip();

                    if (!hatte_schon_subjekt) {
                    	subj += word;
                        subj += " ";

                        if (wa1 == NOUN && (wa_next != LINKING)) {
                        	subj += "{";
                        	subj += adv_bestm;
                        	subj += "} ";
                        	adv_bestm.set("");
                        }
                    }

                    else {
                        obj += word;
                        obj += " ";

                        if (wa1 == NOUN && (wa_next != LINKING)) {
                        	obj += "{";
                        	obj += adv_bestm;
                        	obj += "} ";
                        	adv_bestm.set("");
                        }
                    }

                    if (wa_ == PP) wa_ = NOUN;
                    if (wa_ == NOUN) {
                        hatte_schon_subjekt++;
                    }
                }

            }
        }

        subj.strip();
        prae.strip();
        prae.lower();
        obj.strip();
        questionwort.strip();
        additional_things.strip();
        subj_hinten_dran.strip();

        /*
         * Verben: einheitliche Reihenfolge
         */
        jeliza::strings verben;
        util::split(prae, std::string(" "), verben);
        sort (verben.begin(), verben.end(), pred_lexicographical_compare);
        util::join(prae, std::string(" "), verben);

        jeliza::db_sentence ret;

        ret.subject = subj;
        ret.verb = prae;
        ret.object = obj + " " + subj_hinten_dran;
        ret.object.strip();
        if (!ret.object.size()) {
        	ret.object.set("nothing {}");
        }
        if (!ret.subject.size()) {
        	ret.subject.set("nothing {}");
        }
        ret.quesword = questionwort.get_lower();
        ret.additional = additional_things + adverbiale_bestimmungen;

        if (ret.quesword.size()) {
            ret_list.second.push_back(ret);
        }
        else {
            ret_list.first.push_back(ret);
        }

//        ret.print();

        return ret_list;

    }
    return std::pair<vector<jeliza::db_sentence>, vector<jeliza::db_sentence> >();
}

/*
 * Liest die Wissensdatenbank ein
 */
void jeliza::jeliza_impl::init() {
	log("- Loading database... ");

//    jeliza_db.new_from(("lang_" + jeliza::get_lang_str() + "/jeliza-standard.xml").ref().c_str());
//    jeliza_db.generate_python_db_file();

	log("- Database loaded successful.");
}

jeliza::db_manager& jeliza::jeliza_impl::get_db_manager() {
    return jeliza_db;
}

/*
 * Datenbank neu einlesen von ...
 */
void jeliza::db_manager::new_from(std::string file) {
    m_db.reset(parse(file));
}

/*
 * Lernt einen Satz, dh. formt ihn um (ich->du) und speichert die Datenbank
 */
void jeliza::jeliza_impl::learn (jeliza::strings anss) {
    if (force_answer_check) {
        log("Kann Satz nicht speichern wegen unbekanntem Wort.");
        return;
    }

	init();

	vector<jeliza::db_sentence> into_online_db;

    for (jeliza::strings::iterator it = anss.begin(); it != anss.end(); it++) {
        if (!offline_mode) {
            jeliza::async_download(online_db_new_sentence_entry_url.ref() + url_encode(KI_NAME + ": ").ref() + url_encode(*it).ref());
        }
        if (is_question(*it) != 0) {
            continue;
        }

        log("Speichere " + it->ref());
        it->set(util::transformation(*it));

        jeliza::strings temp;
        util::split(*it, std::string(" "), temp);

        if (temp.size() < 2) {
            log("- Satz hat zu wenig Woerter (< 2): " + it->ref());
            return;
        }

        shared_ptr<jeliza::db_sentence> dbs_ = jeliza::spo_split::get().get_spo(*it);
        jeliza::db_sentence dbs = *dbs_;

//        jeliza_db.push(dbs);
        into_online_db.push_back(dbs);
    }
    log("Saving Database...");
//    jeliza_db.save(("lang_" + jeliza::get_lang_str() + "/jeliza-standard.xml").ref().c_str());
//    jeliza_db.generate_python_db_file();

    log("Sentence saved.");
//    if (use_online_db) {
        log("Using Online Database...");

        for (vector<jeliza::db_sentence>::iterator it = into_online_db.begin(); it != into_online_db.end(); it++) {
            if (!offline_mode) {
                jeliza::async_download(online_db_new_entry_url.ref() + url_encode(it->to_xml()).ref());
            }
        }
        log("Online Database was downloaded...");
//    }
}

/*
 * Wrapper fuer *Datenbank*->push_back
 */
void jeliza::db_manager::push (db_sentence dbs) {
    m_db->push_back(dbs);
}

/*
 * Lernt einen Satz
 */
void jeliza::jeliza_impl::learn (jeliza::string fra) {
    log("learn (jeliza::string)");
	jeliza::strings anss;
	anss.push_back(fra);
	learn(anss);
}

PyObject* jelizacpp_update_prozent(PyObject *, PyObject *args) {
    float fl;

    if (!PyArg_ParseTuple(args, "f", &fl))
        return NULL;

    jeliza::update_prozent(fl);

    return Py_BuildValue("b", true);
}

PyObject* jelizacpp_say(PyObject *, PyObject *args) {
    char* c;

    if (!PyArg_ParseTuple(args, "s", &c))
        return NULL;

    speak(string(c));

    return Py_BuildValue("b", true);
}

PyObject* jelizacpp_is_question(PyObject *, PyObject *args) {
    char* c;

    if (!PyArg_ParseTuple(args, "s", &c))
        return NULL;

    return Py_BuildValue("b", jeliza::jeliza_impl::is_question(s(c)));
}

PyObject* jelizacpp_send(PyObject *, PyObject *args) {
    char* c;

    if (!PyArg_ParseTuple(args, "s", &c))
        return NULL;

    jeliza::comm_send(s(c));

    return Py_BuildValue("");
}

PyMethodDef jelizacpp_methods[] = {
    {"cpp_to_db_sentence",  jelizacpp_to_db_sentence, METH_VARARGS, "Converts a sentence to a 'db_sentence'."},
    {"cpp_has_same_word_base",  jelizacpp_has_same_word_base, METH_VARARGS, "Compares the 'Wortstamm' of two words."},
    {"cpp_update_prozent",  jelizacpp_update_prozent, METH_VARARGS, "Updates the procent label in the main window."},
    {"log",  jelizacpp_log, METH_VARARGS, "Replacement for print which prints the std::string in the log window, too."},
    {"cpp_www_wikipedia_de",  jelizacpp_search_in_wikipedia_with_newlines, METH_VARARGS, "Searches in Wikipedia."},
    {"cpp_listen_link_urls_get",  jelizacpp_listen_link_urls_get, METH_VARARGS, ""},
    {"cpp_detect_wordtype_wrapper",  jelizacpp_detect_word_type_wrapper, METH_VARARGS, ""},
    {"cpp_detect_wordtype",  jelizacpp_detect_word_type, METH_VARARGS, ""},
    {"cpp_transformation", jelizacpp_transformation, METH_VARARGS, ""},
    {"cpp_learn", jelizacpp_learn, METH_VARARGS, ""},
    {"display", jelizacpp_display, METH_VARARGS, ""},
    {"get_force_answer_word", jelizacpp_force_answer_word, METH_VARARGS, ""},
    {"get_force_answer", jelizacpp_force_answer, METH_VARARGS, ""},
    {"add_word", jelizacpp_add_word, METH_VARARGS, ""},
    {"cpp_say", jelizacpp_say, METH_VARARGS, ""},
    {"cpp_is_question", jelizacpp_is_question, METH_VARARGS, ""},
    {"send", jelizacpp_send, METH_VARARGS, ""},

    {NULL, NULL, 0, NULL}
};

PyObject* jelizacpp_listen_link_urls_get(PyObject *, PyObject *args) {
    unsigned int index = 1;

    if (!PyArg_ParseTuple(args, "i", &index))
        return NULL;

    index--;
    jeliza::string def;
    if (index < listen_link_urls.size()) {
        std::string res = listen_link_urls[index].ref();
        log("Resultierender Link:" + res);
        def = search_in_wikipedia_with_newlines(jeliza::string(res));
        log("Result:" + def.ref());
    }

    return Py_BuildValue("s", def.ref().c_str());
}

PyObject* jelizacpp_detect_word_type_wrapper (PyObject *, PyObject *args) {
    unsigned int wordtype;
    const char *command;
    bool atBegin = false;

    if (!PyArg_ParseTuple(args, "sb", &command, &atBegin))
        return NULL;

    wordtype = jeliza::spo_split::get().detect_word_type_wrapper(jeliza::string(string(command)), atBegin);

    return Py_BuildValue("i", wordtype);
}

PyObject* jelizacpp_detect_word_type (PyObject *, PyObject *args) {
    unsigned int wordtype;
    const char *command;
    bool atBegin = false;

    if (!PyArg_ParseTuple(args, "sb", &command, &atBegin))
        return NULL;

    wordtype = jeliza::spo_split::get().detect_word_type(jeliza::string(string(command)), atBegin);

    return Py_BuildValue("i", wordtype);
}

PyObject* jelizacpp_search_in_wikipedia_with_newlines(PyObject *, PyObject *args) {
    const char *command;

    if (!PyArg_ParseTuple(args, "s", &command))
        return NULL;

    jeliza::string def = search_in_wikipedia_with_newlines(jeliza::string(command));
    log("Result:" + def.ref());

    return Py_BuildValue("s", def.ref().c_str());
}



PyObject* jelizacpp_to_db_sentence(PyObject *, PyObject *args) {
    const char *command;

    if (!PyArg_ParseTuple(args, "s", &command))
        return NULL;

    shared_ptr<jeliza::db_sentence> dbs = jeliza::spo_split::get().get_spo(jeliza::string(string(command)));
//    dbs->print();

    jeliza::db_sentence dbss;
    jeliza::db_sentence dbss2;
    jeliza::db_sentence dbss3;
    jeliza::db_sentence dbss4;
    jeliza::db_sentence dbss5;

    if (dbs->sub->size() > 0) {
        if (&(*dbs->sub)[0]) {
            dbss = (*dbs->sub)[0];
        }
    }
    if (dbs->sub->size() > 1) {
        if (&(*dbs->sub)[1]) {
            dbss2 = (*dbs->sub)[1];
        }
    }
    if (dbs->sub->size() > 2) {
        if (&(*dbs->sub)[2]) {
            dbss3 = (*dbs->sub)[2];
        }
    }
    if (dbs->sub->size() > 3) {
        if (&(*dbs->sub)[3]) {
            dbss4 = (*dbs->sub)[3];
        }
    }
    if (dbs->sub->size() > 4) {
        if (&(*dbs->sub)[4]) {
            dbss5 = (*dbs->sub)[4];
        }
    }

//    dbs->print();

    return Py_BuildValue
            (
            "(sssssssissbsssssssissbsssssssissbsssssssissbsssssssissbsssssssiss)",

            dbs->subject.get().c_str(), dbs->verb.get().c_str(),
            dbs->object.get().c_str(), dbs->prefix.get().c_str(), dbs->suffix.get().c_str(),
            dbs->feeling.get().c_str(), dbs->category.get().c_str(), dbs->priority, dbs->additional.get().c_str(),
            dbs->quesword.get().c_str(),

            dbs->sub->size() > 0, dbss.subject.get().c_str(), dbss.verb.get().c_str(),
            dbss.object.get().c_str(), dbss.prefix.get().c_str(), dbss.suffix.get().c_str(),
            dbss.feeling.get().c_str(), dbss.category.get().c_str(), dbss.priority, dbss.additional.get().c_str(),
            dbss.quesword.get().c_str(),

            dbs->sub->size() > 1, dbss2.subject.get().c_str(), dbss2.verb.get().c_str(),
            dbss2.object.get().c_str(), dbss2.prefix.get().c_str(), dbss2.suffix.get().c_str(),
            dbss2.feeling.get().c_str(), dbss2.category.get().c_str(), dbss2.priority, dbss2.additional.get().c_str(),
            dbss2.quesword.get().c_str(),

            dbs->sub->size() > 2, dbss3.subject.get().c_str(), dbss3.verb.get().c_str(),
            dbss3.object.get().c_str(), dbss3.prefix.get().c_str(), dbss3.suffix.get().c_str(),
            dbss3.feeling.get().c_str(), dbss3.category.get().c_str(), dbss3.priority, dbss3.additional.get().c_str(),
            dbss3.quesword.get().c_str(),

            dbs->sub->size() > 3, dbss4.subject.get().c_str(), dbss4.verb.get().c_str(),
            dbss4.object.get().c_str(), dbss4.prefix.get().c_str(), dbss4.suffix.get().c_str(),
            dbss4.feeling.get().c_str(), dbss4.category.get().c_str(), dbss4.priority, dbss4.additional.get().c_str(),
            dbss4.quesword.get().c_str(),

            dbs->sub->size() > 4, dbss5.subject.get().c_str(), dbss5.verb.get().c_str(),
            dbss5.object.get().c_str(), dbss5.prefix.get().c_str(), dbss5.suffix.get().c_str(),
            dbss5.feeling.get().c_str(), dbss5.category.get().c_str(), dbss5.priority, dbss5.additional.get().c_str(),
            dbss5.quesword.get().c_str()

            );
}

PyObject* jelizacpp_log(PyObject *, PyObject *args) {
    const char *command;

    if (!PyArg_ParseTuple(args, "s", &command))
        return NULL;

    jeliza::log(string(command));

    return Py_BuildValue("s", command);
}

PyObject* jelizacpp_display(PyObject *, PyObject *args) {
    const char *command;

    if (!PyArg_ParseTuple(args, "s", &command))
        return NULL;

    jeliza::display_sentence(jeliza::string(string(command)).get_strip());

    return Py_BuildValue("b", 0);
}

PyObject* jelizacpp_transformation(PyObject *, PyObject *args) {
    const char *command;

    if (!PyArg_ParseTuple(args, "s", &command))
        return NULL;

    jeliza::string s;
    s.set(string(command));
    s = util::transformation(s);

    return Py_BuildValue("s", s.ref().c_str());
}
PyObject* jelizacpp_learn(PyObject *, PyObject *args) {
    const char *command;

    if (!PyArg_ParseTuple(args, "s", &command))
        return NULL;

    jeliza::string s;
    s.set(string(command));
    jeliza_impl_instance.learn(s);

    return Py_BuildValue("s", s.ref().c_str());
}

PyObject* jelizacpp_force_answer_word(PyObject *, PyObject *) {
    return Py_BuildValue("s", jeliza::force_answer_word.c_str());
}

PyObject* jelizacpp_force_answer(PyObject *, PyObject *) {
    return Py_BuildValue("s", jeliza::force_answer.ref().c_str());
}

PyObject* jelizacpp_add_word(PyObject *, PyObject *args) {
    const char *command;

    if (!PyArg_ParseTuple(args, "s", &command))
        return NULL;

    jeliza::jeliza_manager::get_current()->get().add_word(command[0]);

    return Py_BuildValue("b", false);
}



PyObject* jelizacpp_has_same_word_base(PyObject*, PyObject* args) {
    const char *s1;
    const char *s2;

    if (!PyArg_ParseTuple(args, "ss", &s1, &s2))
        return NULL;

    jeliza::string stamm1 = jeliza::spo_split::get().detect_word_base(jeliza::string(s1));
    jeliza::string stamm2 = jeliza::spo_split::get().detect_word_base(jeliza::string(s2));

//    jeliza::string prefixe_entfernt1 = jeliza::spo_split::get().delete_prefixes(jeliza::string(s1));
//    jeliza::string prefixe_entfernt2 = jeliza::spo_split::get().delete_prefixes(jeliza::string(s2));

    bool hat_gleichen_stamm = jeliza::spo_split::get().has_same_word_base(jeliza::string(s1), jeliza::string(s2),
            stamm1, stamm2);

    return Py_BuildValue("(b)", hat_gleichen_stamm);
}

double aktuelle_procent = 0;
void procent_plus(double p) {
    aktuelle_procent += p;

    jeliza::update_prozent(aktuelle_procent);
}

jeliza::DB jeliza::db_manager::get() {
    return *m_db;
}

/*
 * Formt die Datenbank ins Python-Format um und speichert sie
 */
void jeliza::db_manager::generate_python_db_file() {
    ofstream o2("temp/__init__.py");
    o2 << "import pyjdb" << endl;
    o2.close();


    ofstream o("temp/pyjdb.py");


    o << "db = []" << endl;

    for (unsigned int x = 0; x < m_db->size(); ++x) {
        db_sentence* it = new db_sentence((*m_db)[x]);
        it->strip();
        o << endl
          << "tempdbsent" << x << " = db_sentence()" << endl
          << "tempdbsent" << x << ".subject        = '" << it->subject.replace(string("'"), std::string("\"")).ref() << "'" << endl
          << "tempdbsent" << x << ".verb           = '" << it->verb.replace(string("'"), std::string("\"")).ref() << "'" << endl
          << "tempdbsent" << x << ".object         = '" << it->object.replace(string("'"), std::string("\"")).ref() << "'" << endl
          << "tempdbsent" << x << ".prefix         = '" << it->prefix.replace(string("'"), std::string("\"")).ref() << "'" << endl
          << "tempdbsent" << x << ".suffix         = '" << it->suffix.replace(string("'"), std::string("\"")).ref() << "'" << endl
          << "tempdbsent" << x << ".feeling        = '" << it->feeling.replace(string("'"), std::string("\"")).ref() << "'" << endl
          << "tempdbsent" << x << ".category       = '" << it->category.replace(string("'"), std::string("\"")).ref() << "'" << endl
          << "tempdbsent" << x << ".priority       = " << it->priority << "" << endl
          << "tempdbsent" << x << ".additional         = '" << it->additional.replace(string("'"), std::string("\"")).ref() << "'" << endl
          << "tempdbsent" << x << ".quesword       = '" << it->quesword.replace(string("'"), std::string("\"")).ref() << "'" << endl
//          << "tempdbsent" << x << ".all_without_nonsense = '" << jeliza::without_nonsense(it->gen_sentences_all(true)[0]).replace(string("'"), std::string("\"")).ref() << "'" << endl
//          << "tempdbsent" << x << ".sub = db_sentence()" << endl
//          << "tempdbsent" << x << ".has_sub = False" << endl
          << endl;

        o << endl;
        for (unsigned int y = 0; y < it->sub->size(); ++y) {
            o << endl
              << "tempdbsent_sub" << x << "_" << y << " = db_sentence()" << endl
              << "tempdbsent_sub" << x << "_" << y << ".subject        = '" << (*it->sub)[y].subject.replace(string("'"), std::string("\"")).ref() << "'" << endl
              << "tempdbsent_sub" << x << "_" << y << ".verb           = '" << (*it->sub)[y].verb.replace(string("'"), std::string("\"")).ref() << "'" << endl
              << "tempdbsent_sub" << x << "_" << y << ".object         = '" << (*it->sub)[y].object.replace(string("'"), std::string("\"")).ref() << "'" << endl
              << "tempdbsent_sub" << x << "_" << y << ".prefix         = '" << (*it->sub)[y].prefix.replace(string("'"), std::string("\"")).ref() << "'" << endl
              << "tempdbsent_sub" << x << "_" << y << ".suffix         = '" << (*it->sub)[y].suffix.replace(string("'"), std::string("\"")).ref() << "'" << endl
              << "tempdbsent_sub" << x << "_" << y << ".feeling        = '" << (*it->sub)[y].feeling.replace(string("'"), std::string("\"")).ref() << "'" << endl
              << "tempdbsent_sub" << x << "_" << y << ".category       = '" << (*it->sub)[y].category.replace(string("'"), std::string("\"")).ref() << "'" << endl
              << "tempdbsent_sub" << x << "_" << y << ".priority       = " << (*it->sub)[y].priority << "" << endl
              << "tempdbsent_sub" << x << "_" << y << ".additional         = '" << (*it->sub)[y].additional.replace(string("'"), std::string("\"")).ref() << "'" << endl
              << "tempdbsent_sub" << x << "_" << y << ".quesword       = '" << (*it->sub)[y].quesword.replace(string("'"), std::string("\"")).ref() << "'" << endl
//              << "tempdbsent" << x << ".sub.all_without_nonsense = '" << jeliza::without_nonsense((*it->sub)[0].gen_sentences_all(true)[0]).replace(string("'"), std::string("\"")).ref() << "'" << endl
              << "tempdbsent" << x << ".sub" << y+1 << " = " << "tempdbsent_sub" << x << "_" << y << "" << endl
              << "tempdbsent" << x << ".has" << y+1 << "_sub = True" << endl
              << endl;
        }
        o << endl
          << "db.append(tempdbsent" << x << ")" << endl;
    }


    o << endl << endl;
    o.flush();
    o.close();
}


/*
 * answeret auf eine question
 */
void jeliza::jeliza_impl::ask(jeliza::string question) {
	clogger << "- Suche eine passende answer: " << endl;
    procent_plus(1);

    // Clearing the cache which decides whether a word is a noun ot not a noun if it is at the beginning of a sentence
    cache_noun_or_not = map<std::string, jeliza::wordtype>();

    ofstream o("temp/answer.tmp");
    o.close();

    jeliza::string ans;


    ofstream o2("temp/new_sentence.tmp");
    o2 << question.ref();
    o2.close();
}

jeliza::string util::transformation (jeliza::string str) {
	str = str.replace("-", "");
	str = str.replace(";", "");
	str = str.replace("!", ".");
	str = str.replace_nocase_safe(s(" ,"), s(","));
	str = str.replace_nocase_safe(s(", "), s(","));
	str = str.replace_nocase_safe(s(","), s(" , "));
	str = str.replace("?", " ucebckiundihwec");
	str = str.replace("ucebckiundihwec", "?");

	str = util::change(str, s("ich"), s("du"));
	str = util::change(str, s("mein"), s("dein"));
	str = util::change(str, s("meine"), s("deine"));
	str = util::change(str, s("meins"), s("deins"));
	str = util::change(str, s("meiner"), s("deiner"));
	str = util::change(str, s("meinem"), s("deinem"));
	str = util::change(str, s("mir"), s("dir"));
	str = util::change(str, s("mich"), s("dich"));
	str = util::change(str, s("user"), s("euer"));
	str = util::change(str, s("usere"), s("euere"));
	str = util::change(str, s("userer"), s("eurer"));
	str = util::change(str, s("useres"), s("euers"));

	str = util::change(str, s("am"), s("are"));

	str = str.replace("wieso", "warum");
	str = str.replace("weshalb", "warum");
	str = (" " + str + " ").replace_nocase_safe(s("whom"), s("who")).get_strip();

	ifstream in("verbs.txt");
	vector<std::string> verbs;
	jeliza::string buffer;
	while (in) {
		getline(in, buffer.ref());
		buffer.strip();
		verbs.push_back(buffer.ref());
	}

	vector<std::string> ss;
	jeliza::strings ss2;
	split(str, std::string(" "), ss2);
	jeliza::string rep;

    for (jeliza::strings::iterator it = ss2.begin(); it != ss2.end(); it++) {
        ss.push_back(it->ref());
    }

    unsigned int noun_count = 0;
	for (unsigned int y = 0; y < ss.size(); y++) {
	    jeliza::string w(ss[y]);
		jeliza::string lower(w);
		lower.lower();
		w.strip();
		lower.strip();
		jeliza::wordtype wa = jeliza::spo_split::get().detect_word_type((y == 0) ? lower : w, y == 0);
		if (wa == jeliza::NOUN) noun_count++;

        bool is_verb = false;
        if (w.size() > 0) {
            is_verb = (wa == jeliza::VERB);
        }

        jeliza::string w2("-" + lower + "-");
        w2 = w2.replace(",-", "-,");
		if (is_verb && w2.contains("st-") && (w == lower || y == 0) && jeliza::get_lang() == jeliza::GERMAN) {
            jeliza::string verb2("-" + lower + "-");
            verb2 = verb2.replace(",-", "-,");
            if (verb2.contains("-bist-")) {
    			rep += "bin";
    			rep += " ";
    			continue;
            }
            if (verb2.contains("-hast-")) {
    			rep += "habe";
    			rep += " ";
    			continue;
            }
            verb2 = verb2.replace("willst", "will");
            verb2 = verb2.replace_nocase_safe(s("sst-"), s("ssst-"));
            verb2 = verb2.replace("st-", "e-");
            verb2 = verb2.replace("zt-", "ze-");
            verb2 = verb2.replace("ee-", "hoire-");
            verb2 = verb2.replace("hoire-", "e-");
            verb2 = verb2.replace("-", "");
        	if (verb2 == "ie") verb2.set("ist");
			rep += verb2.ref();
			rep += " ";
		}
		else if (is_verb && (w == lower || y == 0) && jeliza::get_lang() == jeliza::GERMAN) {
            jeliza::string verb2("-" + lower + "-");
            verb2 = verb2.replace(",-", "-,");
            if (verb2.contains("-bin-")) {
    			rep += "bist";
    			rep += " ";
    			continue;
            }
            if (verb2.contains("-habe-")) {
    			rep += "hast";
    			rep += " ";
    			continue;
            }
            verb2 = verb2.replace_nocase_safe(s("will"), s("willst"));
            verb2 = verb2.replace("nte-", "ntee-");
            verb2 = verb2.replace("lte-", "ltee-");
            verb2 = verb2.replace("mte-", "mtee-");
            verb2 = verb2.replace("rte-", "rtee-");
            verb2 = verb2.replace("pte-", "ptee-");
            verb2 = verb2.replace("bte-", "btee-");
            verb2 = verb2.replace("kte-", "ktee-");
            verb2 = verb2.replace("tte-", "ttee-");
            verb2 = verb2.replace("e-", "st-");
            verb2 = verb2.replace("zst-", "zt-");
            verb2 = verb2.replace("-", "");
			rep += verb2.ref();
			rep += " ";
		}
		else {
			w = w.replace("-", "");
			bool komma = false;
			if (w.contains(s(","))) {
			    komma = true;
			    w = w.replace(",", "");
			}

            if (w.get_lower() == "you" && noun_count == 1) w.set("I");
            else if (w.get_lower() == "you") w.set("me");
            else if (w.get_lower() == "i") w.set("you");
            else if (w.get_lower() == "me") w.set("you");
            else if (w.get_lower() == "my") w.set("your");
            else if (w.get_lower() == "your") w.set("my");
            else if (w.get_lower() == "mine") w.set("yours");
            else if (w.get_lower() == "yours") w.set("mine");
            else if (w.get_lower() == "yourself") w.set("myself");
            else if (w.get_lower() == "myself") w.set("yourself");
            else if (w.get_lower() == "yourselves") w.set("ourselves");
            else if (w.get_lower() == "ourselves") w.set("yourselves");
            else if (w.get_lower() == "our") w.set("your");
            else if (w.get_lower() == "us") w.set("you");

			rep += w.replace("-", "").ref();
			if (komma) {
			    rep += ",";
			    noun_count = 0;
			}
			rep += " ";
		}

	}

    str = rep;
	str = str.replace(string("-"), std::string(""));
	str = str.replace(string(" ,"), std::string(","));
//	str = util::change(str, s("heisse"), s("heisst"));
	str = str.replace(string("eise"), std::string("eisse"));
	str = str.replace(string("wist"), std::string("wie"));
//	str = util::change(str, s("bin"), s("bist"));


//	str = util::change(str, s("habst"), s("hast"));

	return (jeliza::string(str));
}



#endif
